# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Program definition for a distributed layout based on a builder."""

from typing import Any, Callable, Dict, Optional, Union, Sequence

from acme import core
from acme.jax import networks as networks_lib
from acme.jax import savers
from acme.jax import types
from acme.jax import utils
from acme.utils import counting
from acme.utils import loggers
from acme.utils import lp_utils
import jax
import launchpad as lp

AgentNetwork = Any
NetworkFactory = Callable[[], AgentNetwork]
# It will be treated as Dict[str, Any]. Proper support is tracked b/109648354.
NestedLogger = Union[loggers.Logger, Dict[str, 'NestedLogger']]  # pytype: disable=not-supported-yet
LearnerFactory = Callable[[
    networks_lib
    .PRNGKey, AgentNetwork, Optional[counting.Counter], Optional[NestedLogger]
], core.Learner]


class OfflineDistributedLayout:
  """Program definition for an offline distributed agent based on a builder.

  It is distributed in the sense that evaluators run on different machines than
  learner.
  """

  def __init__(
      self,
      seed: int,
      network_factory: NetworkFactory,
      make_learner: LearnerFactory,
      evaluator_factories: Sequence[types.EvaluatorFactory] = (),
      save_logs: bool = False,
      log_every: float = 10.0,
      max_number_of_steps: Optional[int] = None,
      workdir: str = '~/acme',
  ):

    self._seed = seed
    self._make_learner = make_learner
    self._evaluator_factories = evaluator_factories
    self._network_factory = network_factory
    self._save_logs = save_logs
    self._log_every = log_every
    self._max_number_of_steps = max_number_of_steps
    self._workdir = workdir

  def counter(self):
    kwargs = {'directory': self._workdir, 'add_uid': self._workdir == '~/acme'}
    return savers.CheckpointingRunner(
        counting.Counter(), subdirectory='counter', time_delta_minutes=5,
        **kwargs)

  def learner(
      self,
      random_key: networks_lib.PRNGKey,
      counter: counting.Counter,
  ):
    """The Learning part of the agent."""
    # Counter and logger.
    counter = counting.Counter(counter, 'learner')
    logger = loggers.make_default_logger(
        'learner', self._save_logs, time_delta=self._log_every,
        asynchronous=True, serialize_fn=utils.fetch_devicearray,
        steps_key='learner_steps')

    # Create the learner.
    networks = self._network_factory()
    learner = self._make_learner(random_key, networks, counter, logger)

    kwargs = {'directory': self._workdir, 'add_uid': self._workdir == '~/acme'}
    # Return the learning agent.
    return savers.CheckpointingRunner(
        learner, subdirectory='learner', time_delta_minutes=5, **kwargs)

  def coordinator(self, counter: counting.Counter, max_learner_steps: int):
    return lp_utils.StepsLimiter(counter, max_steps=max_learner_steps,
                                 steps_key='learner_steps')

  def build(self, name='agent'):
    """Build the distributed agent topology."""
    program = lp.Program(name=name)

    key = jax.random.PRNGKey(self._seed)

    with program.group('counter'):
      counter = program.add_node(lp.CourierNode(self.counter))
      if self._max_number_of_steps is not None:
        _ = program.add_node(
            lp.CourierNode(self.coordinator, counter,
                           self._max_number_of_steps))

    learner_key, key = jax.random.split(key)
    with program.group('learner'):
      learner = program.add_node(
          lp.CourierNode(self.learner, learner_key, counter))

    with program.group('evaluator'):
      for evaluator in self._evaluator_factories:
        evaluator_key, key = jax.random.split(key)
        program.add_node(
            lp.CourierNode(evaluator, evaluator_key, learner, counter))

    return program



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the single precision wrapper."""

from absl.testing import absltest

from acme import wrappers
from acme.testing import fakes

import numpy as np
import tree


class FakeNonZeroObservationEnvironment(fakes.ContinuousEnvironment):
  """Fake environment with non-zero observations."""

  def _generate_fake_observation(self):
    original_observation = super()._generate_fake_observation()
    return tree.map_structure(np.ones_like, original_observation)


class FrameStackingTest(absltest.TestCase):

  def test_specs(self):
    original_env = FakeNonZeroObservationEnvironment()
    env = wrappers.FrameStackingWrapper(original_env, 2)

    original_observation_spec = original_env.observation_spec()
    expected_shape = original_observation_spec.shape + (2,)
    observation_spec = env.observation_spec()
    self.assertEqual(expected_shape, observation_spec.shape)

    expected_action_spec = original_env.action_spec()
    action_spec = env.action_spec()
    self.assertEqual(expected_action_spec, action_spec)

    expected_reward_spec = original_env.reward_spec()
    reward_spec = env.reward_spec()
    self.assertEqual(expected_reward_spec, reward_spec)

    expected_discount_spec = original_env.discount_spec()
    discount_spec = env.discount_spec()
    self.assertEqual(expected_discount_spec, discount_spec)

  def test_step(self):
    original_env = FakeNonZeroObservationEnvironment()
    env = wrappers.FrameStackingWrapper(original_env, 2)
    observation_spec = env.observation_spec()
    action_spec = env.action_spec()

    timestep = env.reset()
    self.assertEqual(observation_spec.shape, timestep.observation.shape)
    self.assertTrue(np.all(timestep.observation[..., 0] == 0))

    timestep = env.step(action_spec.generate_value())
    self.assertEqual(observation_spec.shape, timestep.observation.shape)

  def test_second_reset(self):
    original_env = FakeNonZeroObservationEnvironment()
    env = wrappers.FrameStackingWrapper(original_env, 2)
    action_spec = env.action_spec()

    env.reset()
    env.step(action_spec.generate_value())
    timestep = env.reset()
    self.assertTrue(np.all(timestep.observation[..., 0] == 0))


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Multihead networks apply separate networks to the input."""

from typing import Callable, Union, Sequence

from acme import types

import sonnet as snt
import tensorflow as tf
import tensorflow_probability as tfp

tfd = tfp.distributions
TensorTransformation = Union[snt.Module, Callable[[types.NestedTensor],
                                                  tf.Tensor]]


class Multihead(snt.Module):
  """Multi-head network module.

  This takes as input a list of N `network_heads`, and returns another network
  whose output is the stacked outputs of each of these network heads separately
  applied to the module input. The dimension of the output is [..., N].
  """

  def __init__(self,
               network_heads: Sequence[TensorTransformation]):
    if not network_heads:
      raise ValueError('Must specify non-empty, non-None critic_network_heads.')
    self._network_heads = network_heads
    super().__init__(name='multihead')

  def __call__(self,
               inputs: tf.Tensor) -> Union[tf.Tensor, Sequence[tf.Tensor]]:
    outputs = [network_head(inputs) for network_head in self._network_heads]
    if isinstance(outputs[0], tfd.Distribution):
      # Cannot stack distributions
      return outputs
    outputs = tf.stack(outputs, axis=-1)
    return outputs



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Environment wrapper that converts environments to use canonical action specs.

This only affects action specs of type `specs.BoundedArray`.

For bounded action specs, we refer to a canonical action spec as the bounding
box [-1, 1]^d where d is the dimensionality of the spec. So the shape and dtype
of the spec is unchanged, while the maximum/minimum values are set to +/- 1.
"""

from acme import specs
from acme import types
from acme.wrappers import base

import dm_env
import numpy as np
import tree


class CanonicalSpecWrapper(base.EnvironmentWrapper):
  """Wrapper which converts environments to use canonical action specs.

  This only affects action specs of type `specs.BoundedArray`.

  For bounded action specs, we refer to a canonical action spec as the bounding
  box [-1, 1]^d where d is the dimensionality of the spec. So the shape and
  dtype of the spec is unchanged, while the maximum/minimum values are set
  to +/- 1.
  """

  def __init__(self, environment: dm_env.Environment, clip: bool = False):
    super().__init__(environment)
    self._action_spec = environment.action_spec()
    self._clip = clip

  def step(self, action: types.NestedArray) -> dm_env.TimeStep:
    scaled_action = _scale_nested_action(action, self._action_spec, self._clip)
    return self._environment.step(scaled_action)

  def action_spec(self):
    return _convert_spec(self._environment.action_spec())


def _convert_spec(nested_spec: types.NestedSpec) -> types.NestedSpec:
  """Converts all bounded specs in nested spec to the canonical scale."""

  def _convert_single_spec(spec: specs.Array) -> specs.Array:
    """Converts a single spec to canonical if bounded."""
    if isinstance(spec, specs.BoundedArray):
      return spec.replace(
          minimum=-np.ones(spec.shape), maximum=np.ones(spec.shape))
    else:
      return spec

  return tree.map_structure(_convert_single_spec, nested_spec)


def _scale_nested_action(
    nested_action: types.NestedArray,
    nested_spec: types.NestedSpec,
    clip: bool,
) -> types.NestedArray:
  """Converts a canonical nested action back to the given nested action spec."""

  def _scale_action(action: np.ndarray, spec: specs.Array):
    """Converts a single canonical action back to the given action spec."""
    if isinstance(spec, specs.BoundedArray):
      # Get scale and offset of output action spec.
      scale = spec.maximum - spec.minimum
      offset = spec.minimum

      # Maybe clip the action.
      if clip:
        action = np.clip(action, -1.0, 1.0)

      # Map action to [0, 1].
      action = 0.5 * (action + 1.0)

      # Map action to [spec.minimum, spec.maximum].
      action *= scale
      action += offset

    return action

  return tree.map_structure(_scale_action, nested_action, nested_spec)



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for SVG0 algorithm with priors."""

import collections
from typing import Tuple, Optional, Dict, Iterable

from acme import types
from acme.tf import utils as tf2_utils

import sonnet as snt
import tensorflow as tf
import tree


class OnlineTargetPiQ(snt.Module):
  """Core to unroll online and target policies and Q functions at once.

  A core that runs online and target policies and Q functions. This can be more
  efficient if the core needs to be unrolled across time and called many times.
  """

  def __init__(self,
               online_pi: snt.Module,
               online_q: snt.Module,
               target_pi: snt.Module,
               target_q: snt.Module,
               num_samples: int,
               online_prior: Optional[snt.Module] = None,
               target_prior: Optional[snt.Module] = None,
               name='OnlineTargetPiQ'):
    super().__init__(name)

    self._online_pi = online_pi
    self._target_pi = target_pi
    self._online_q = online_q
    self._target_q = target_q
    self._online_prior = online_prior
    self._target_prior = target_prior

    self._num_samples = num_samples
    output_list = [
        'online_samples', 'target_samples', 'target_log_probs_behavior_actions',
        'online_log_probs', 'online_q', 'target_q'
    ]
    if online_prior is not None:
      output_list += ['analytic_kl_divergence', 'analytic_kl_to_target']
    self._output_tuple = collections.namedtuple(
        'OnlineTargetPiQ', output_list)

  def __call__(self, input_obs_and_action: Tuple[tf.Tensor, tf.Tensor]):
    (obs, action) = input_obs_and_action
    online_pi_dist = self._online_pi(obs)
    target_pi_dist = self._target_pi(obs)

    online_samples = online_pi_dist.sample(self._num_samples)
    target_samples = target_pi_dist.sample(self._num_samples)
    target_log_probs_behavior_actions = target_pi_dist.log_prob(action)

    online_log_probs = online_pi_dist.log_prob(tf.stop_gradient(online_samples))

    online_q_out = self._online_q(obs, action)
    target_q_out = self._target_q(obs, action)

    output_list = [
        online_samples, target_samples, target_log_probs_behavior_actions,
        online_log_probs, online_q_out, target_q_out
    ]

    if self._online_prior is not None:
      prior_dist = self._online_prior(obs)
      target_prior_dist = self._target_prior(obs)
      analytic_kl_divergence = online_pi_dist.kl_divergence(prior_dist)
      analytic_kl_to_target = online_pi_dist.kl_divergence(target_prior_dist)

      output_list += [analytic_kl_divergence, analytic_kl_to_target]
    output = self._output_tuple(*output_list)
    return output


def static_rnn(core: snt.Module, inputs: types.NestedTensor,
               unroll_length: int):
  """Unroll core along inputs for unroll_length steps.

  Note: for time-major input tensors whose leading dimension is less than
  unroll_length, `None` would be provided instead.

  Args:
    core: an instance of snt.Module.
    inputs: a `nest` of time-major input tensors.
    unroll_length: number of time steps to unroll.

  Returns:
    step_outputs: a `nest` of time-major stacked output tensors of length
      `unroll_length`.
  """
  step_outputs = []
  for time_dim in range(unroll_length):
    inputs_t = tree.map_structure(
        lambda t, i_=time_dim: t[i_] if i_ < t.shape[0] else None, inputs)
    step_output = core(inputs_t)
    step_outputs.append(step_output)

  step_outputs = _nest_stack(step_outputs)
  return step_outputs


def mask_out_restarting(tensor: tf.Tensor, start_of_episode: tf.Tensor):
  """Mask out `tensor` taken on the step that resets the environment.

  Args:
    tensor: a time-major 2-D `Tensor` of shape [T, B].
    start_of_episode: a 2-D `Tensor` of shape [T, B] that contains the points
      where the episode restarts.

  Returns:
    tensor of shape [T, B] with elements are masked out according to step_types,
    restarting weights of shape [T, B]
  """
  tensor.get_shape().assert_has_rank(2)
  start_of_episode.get_shape().assert_has_rank(2)
  weights = tf.cast(~start_of_episode, dtype=tf.float32)
  masked_tensor = tensor * weights
  return masked_tensor


def batch_concat_selection(observation_dict: Dict[str, types.NestedTensor],
                           concat_keys: Optional[Iterable[str]] = None,
                           output_dtype=tf.float32) -> tf.Tensor:
  """Concatenate a dict of observations into 2-D tensors."""
  concat_keys = concat_keys or sorted(observation_dict.keys())
  to_concat = []
  for obs in concat_keys:
    if obs not in observation_dict:
      raise KeyError(
          'Missing observation. Requested: {} (available: {})'.format(
              obs, list(observation_dict.keys())))
    to_concat.append(tf.cast(observation_dict[obs], output_dtype))

  return tf2_utils.batch_concat(to_concat)


def _nest_stack(list_of_nests, axis=0):
  """Convert a list of nests to a nest of stacked lists."""
  return tree.map_structure(lambda *ts: tf.stack(ts, axis=axis), *list_of_nests)



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DQN actor helpers."""

from typing import Callable, Sequence

from acme.agents.jax import actor_core as actor_core_lib
from acme.jax import networks as networks_lib
from acme.jax import utils
import chex
import jax
import jax.numpy as jnp
import rlax


Epsilon = float
EpsilonPolicy = Callable[[
    networks_lib.Params, networks_lib.PRNGKey, networks_lib
    .Observation, Epsilon
], networks_lib.Action]


@chex.dataclass(frozen=True, mappable_dataclass=False)
class EpsilonActorState:
  rng: networks_lib.PRNGKey
  epsilon: jnp.ndarray


def alternating_epsilons_actor_core(
    policy_network: EpsilonPolicy, epsilons: Sequence[float],
) -> actor_core_lib.ActorCore[EpsilonActorState, None]:
  """Returns actor components for alternating epsilon exploration.

  Args:
    policy_network: A feedforward action selecting function.
    epsilons: epsilons to alternate per-episode for epsilon-greedy exploration.

  Returns:
    A feedforward policy.
  """
  epsilons = jnp.array(epsilons)

  def apply_and_sample(params: networks_lib.Params,
                       observation: networks_lib.Observation,
                       state: EpsilonActorState):
    random_key, key = jax.random.split(state.rng)
    actions = policy_network(params, key, observation, state.epsilon)
    return (actions.astype(jnp.int32),
            EpsilonActorState(rng=random_key, epsilon=state.epsilon))

  def policy_init(random_key: networks_lib.PRNGKey):
    random_key, key = jax.random.split(random_key)
    epsilon = jax.random.choice(key, epsilons)
    return EpsilonActorState(rng=random_key, epsilon=epsilon)

  return actor_core_lib.ActorCore(
      init=policy_init, select_action=apply_and_sample,
      get_extras=lambda _: None)


def behavior_policy(network: networks_lib.FeedForwardNetwork
                    ) -> EpsilonPolicy:
  """A policy with parameterized epsilon-greedy exploration."""

  def apply_and_sample(params: networks_lib.Params, key: networks_lib.PRNGKey,
                       observation: networks_lib.Observation, epsilon: Epsilon
                       ) -> networks_lib.Action:
    # TODO(b/161332815): Make JAX Actor work with batched or unbatched inputs.
    observation = utils.add_batch_dim(observation)
    action_values = network.apply(params, observation)
    action_values = utils.squeeze_batch_dim(action_values)
    return rlax.epsilon_greedy(epsilon).sample(key, action_values)

  return apply_and_sample


def default_behavior_policy(network: networks_lib.FeedForwardNetwork,
                            epsilon: Epsilon) -> EpsilonPolicy:
  """A policy with a fixed-epsilon epsilon-greedy exploration.

  DEPRECATED: use behavior_policy instead.
  Args:
    network: network producing observation -> action values or logits
    epsilon: sampling parameter that overrides the one in EpsilonPolicy
  Returns:
    epsilon-greedy behavior policy with fixed epsilon
  """
  # TODO(lukstafi): remove this function and migrate its users.

  def apply_and_sample(params: networks_lib.Params, key: networks_lib.PRNGKey,
                       observation: networks_lib.Observation, _: Epsilon
                       ) -> networks_lib.Action:
    # TODO(b/161332815): Make JAX Actor work with batched or unbatched inputs.
    observation = utils.add_batch_dim(observation)
    action_values = network.apply(params, observation)
    action_values = utils.squeeze_batch_dim(action_values)
    return rlax.epsilon_greedy(epsilon).sample(key, action_values)

  return apply_and_sample



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A library of useful adder wrappers."""

from typing import Iterable

from acme import types
from acme.adders import base
import dm_env


class ForkingAdder(base.Adder):
  """An adder that forks data into several other adders."""

  def __init__(self, adders: Iterable[base.Adder]):
    self._adders = adders

  def reset(self):
    for adder in self._adders:
      adder.reset()

  def add_first(self, timestep: dm_env.TimeStep):
    for adder in self._adders:
      adder.add_first(timestep)

  def add(self,
          action: types.NestedArray,
          next_timestep: dm_env.TimeStep,
          extras: types.NestedArray = ()):
    for adder in self._adders:
      adder.add(action, next_timestep, extras)


class IgnoreExtrasAdder(base.Adder):
  """An adder that ignores extras."""

  def __init__(self, adder: base.Adder):
    self._adder = adder

  def reset(self):
    self._adder.reset()

  def add_first(self, timestep: dm_env.TimeStep):
    self._adder.add_first(timestep)

  def add(self,
          action: types.NestedArray,
          next_timestep: dm_env.TimeStep,
          extras: types.NestedArray = ()):
    self._adder.add(action, next_timestep)



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common JAX type definitions."""

from typing import Callable, Generic, Mapping, TypeVar

from acme import core
from acme import types
from acme.utils import counting
import chex
import jax.numpy as jnp

PRNGKey = jnp.ndarray
EvaluatorFactory = Callable[[PRNGKey, core.VariableSource, counting.Counter],
                            core.Worker]
Networks = TypeVar('Networks')
PolicyNetwork = TypeVar('PolicyNetwork')
Sample = TypeVar('Sample')
TrainingState = TypeVar('TrainingState')

TrainingMetrics = Mapping[str, jnp.ndarray]
"""Metrics returned by the training step.

Typically these are logged, so the values are expected to be scalars.
"""

Variables = Mapping[str, types.NestedArray]
"""Mapping of variable collections.

A mapping of variable collections, as defined by Learner.get_variables.
The keys are the collection names, the values are nested arrays representing
the values of the corresponding collection variables.
"""


@chex.dataclass(frozen=True, mappable_dataclass=False)
class TrainingStepOutput(Generic[TrainingState]):
  state: TrainingState
  metrics: TrainingMetrics



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for logging filters."""

import time

from absl.testing import absltest
from acme.utils.loggers import base
from acme.utils.loggers import filters


# TODO(jaslanides): extract this to test_utils, or similar, for re-use.
class FakeLogger(base.Logger):
  """A fake logger for testing."""

  def __init__(self):
    self.data = []

  def write(self, data):
    self.data.append(data)

  @property
  def last_write(self):
    return self.data[-1]

  def close(self):
    pass


class GatedFilterTest(absltest.TestCase):

  def test_logarithmic_filter(self):
    logger = FakeLogger()
    filtered = filters.GatedFilter.logarithmic(logger, n=10)
    for t in range(100):
      filtered.write({'t': t})
    rows = [row['t'] for row in logger.data]
    self.assertEqual(rows, [*range(10), *range(10, 100, 10)])

  def test_periodic_filter(self):
    logger = FakeLogger()
    filtered = filters.GatedFilter.periodic(logger, interval=10)
    for t in range(100):
      filtered.write({'t': t})
    rows = [row['t'] for row in logger.data]
    self.assertEqual(rows, list(range(0, 100, 10)))


class TimeFilterTest(absltest.TestCase):

  def test_delta(self):
    logger = FakeLogger()
    filtered = filters.TimeFilter(logger, time_delta=0.1)

    # Logged.
    filtered.write({'foo': 1})
    self.assertIn('foo', logger.last_write)

    # *Not* logged.
    filtered.write({'bar': 2})
    self.assertNotIn('bar', logger.last_write)

    # Wait out delta.
    time.sleep(0.11)

    # Logged.
    filtered.write({'baz': 3})
    self.assertIn('baz', logger.last_write)

    self.assertLen(logger.data, 2)


class KeyFilterTest(absltest.TestCase):

  def test_keep_filter(self):
    logger = FakeLogger()
    filtered = filters.KeyFilter(logger, keep=('foo',))
    filtered.write({'foo': 'bar', 'baz': 12})
    row, *_ = logger.data
    self.assertIn('foo', row)
    self.assertNotIn('baz', row)

  def test_drop_filter(self):
    logger = FakeLogger()
    filtered = filters.KeyFilter(logger, drop=('foo',))
    filtered.write({'foo': 'bar', 'baz': 12})
    row, *_ = logger.data
    self.assertIn('baz', row)
    self.assertNotIn('foo', row)

  def test_bad_arguments(self):
    with self.assertRaises(ValueError):
      filters.KeyFilter(FakeLogger())
    with self.assertRaises(ValueError):
      filters.KeyFilter(FakeLogger(), keep=('a',), drop=('b',))


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SAC config."""
import dataclasses
from typing import Any, Optional

from acme import specs
from acme.adders import reverb as adders_reverb
import numpy as onp


@dataclasses.dataclass
class SACConfig:
  """Configuration options for SAC."""
  # Loss options
  batch_size: int = 256
  learning_rate: float = 3e-4
  reward_scale: float = 1
  discount: float = 0.99
  n_step: int = 1
  # Coefficient applied to the entropy bonus. If None, an adaptative
  # coefficient will be used.
  entropy_coefficient: Optional[float] = None
  target_entropy: float = 0.0
  # Target smoothing coefficient.
  tau: float = 0.005

  # Replay options
  min_replay_size: int = 10000
  max_replay_size: int = 1000000
  replay_table_name: str = adders_reverb.DEFAULT_PRIORITY_TABLE
  prefetch_size: int = 4
  samples_per_insert: float = 256
  # Rate to be used for the SampleToInsertRatio rate limitter tolerance.
  # See a formula in make_replay_tables for more details.
  samples_per_insert_tolerance_rate: float = 0.1

  # How many gradient updates to perform per step.
  num_sgd_steps_per_step: int = 1


def target_entropy_from_env_spec(
    spec: specs.EnvironmentSpec,
    target_entropy_per_dimension: Optional[float] = None,
) -> float:
  """A heuristic to determine a target entropy.

  If target_entropy_per_dimension is not specified, the target entropy is
  computed as "-num_actions", otherwise it is
  "target_entropy_per_dimension * num_actions".

  Args:
    spec: environment spec
    target_entropy_per_dimension: None or target entropy per action dimension

  Returns:
    target entropy
  """

  def get_num_actions(action_spec: Any) -> float:
    """Returns a number of actions in the spec."""
    if isinstance(action_spec, specs.BoundedArray):
      return onp.prod(action_spec.shape, dtype=int)
    elif isinstance(action_spec, tuple):
      return sum(get_num_actions(subspace) for subspace in action_spec)
    else:
      raise ValueError('Unknown action space type.')

  num_actions = get_num_actions(spec.actions)
  if target_entropy_per_dimension is None:
    if not isinstance(spec.actions, specs.BoundedArray) or isinstance(
        spec.actions, specs.DiscreteArray):
      raise ValueError('Only accept BoundedArrays for automatic '
                       f'target_entropy, got: {spec.actions}')
    if not onp.all(spec.actions.minimum == -1.):
      raise ValueError(
          f'Minimum expected to be -1, got: {spec.actions.minimum}')
    if not onp.all(spec.actions.maximum == 1.):
      raise ValueError(
          f'Maximum expected to be 1, got: {spec.actions.maximum}')

    return -num_actions
  else:
    return target_entropy_per_dimension * num_actions



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Discrete BCQ learner implementation.

As described in https://arxiv.org/pdf/1910.01708.pdf.
"""

import copy
from typing import Dict, List, Optional

from acme import core
from acme import types
from acme.adders import reverb as adders
from acme.agents.tf import bc
from acme.tf import losses
from acme.tf import savers as tf2_savers
from acme.tf import utils as tf2_utils
from acme.tf.networks import discrete as discrete_networks
from acme.utils import counting
from acme.utils import loggers
import numpy as np
import reverb
import sonnet as snt
import tensorflow as tf
import trfl


class _InternalBCQLearner(core.Learner, tf2_savers.TFSaveable):
  """Internal BCQ learner.

  This implements the Q-learning component in the discrete BCQ algorithm.
  """

  def __init__(
      self,
      network: discrete_networks.DiscreteFilteredQNetwork,
      discount: float,
      importance_sampling_exponent: float,
      learning_rate: float,
      target_update_period: int,
      dataset: tf.data.Dataset,
      huber_loss_parameter: float = 1.,
      replay_client: Optional[reverb.TFClient] = None,
      counter: Optional[counting.Counter] = None,
      logger: Optional[loggers.Logger] = None,
      checkpoint: bool = False,
  ):
    """Initializes the learner.

    Args:
      network: BCQ network
      discount: discount to use for TD updates.
      importance_sampling_exponent: power to which importance weights are raised
        before normalizing.
      learning_rate: learning rate for the q-network update.
      target_update_period: number of learner steps to perform before updating
        the target networks.
      dataset: dataset to learn from, whether fixed or from a replay buffer (see
        `acme.datasets.reverb.make_dataset` documentation).
      huber_loss_parameter: Quadratic-linear boundary for Huber loss.
      replay_client: client to replay to allow for updating priorities.
      counter: Counter object for (potentially distributed) counting.
      logger: Logger object for writing logs to.
      checkpoint: boolean indicating whether to checkpoint the learner.
    """

    # Internalise agent components (replay buffer, networks, optimizer).
    # TODO(b/155086959): Fix type stubs and remove.
    self._iterator = iter(dataset)  # pytype: disable=wrong-arg-types
    self._network = network
    self._q_network = network.q_network
    self._target_q_network = copy.deepcopy(network.q_network)
    self._optimizer = snt.optimizers.Adam(learning_rate)
    self._replay_client = replay_client

    # Internalise the hyperparameters.
    self._discount = discount
    self._target_update_period = target_update_period
    self._importance_sampling_exponent = importance_sampling_exponent
    self._huber_loss_parameter = huber_loss_parameter

    # Learner state.
    self._variables = [self._network.trainable_variables]
    self._num_steps = tf.Variable(0, dtype=tf.int32)

    # Internalise logging/counting objects.
    self._counter = counter or counting.Counter()
    self._logger = logger or loggers.make_default_logger('learner',
                                                         save_data=False)

    # Create a snapshotter object.
    if checkpoint:
      self._snapshotter = tf2_savers.Snapshotter(
          objects_to_save={'network': network}, time_delta_minutes=60.)
    else:
      self._snapshotter = None

  @tf.function
  def _step(self) -> Dict[str, tf.Tensor]:
    """Do a step of SGD and update the priorities."""

    # Pull out the data needed for updates/priorities.
    inputs = next(self._iterator)
    transitions: types.Transition = inputs.data
    keys, probs = inputs.info[:2]

    with tf.GradientTape() as tape:
      # Evaluate our networks.
      q_tm1 = self._q_network(transitions.observation)
      q_t_value = self._target_q_network(transitions.next_observation)
      q_t_selector = self._network(transitions.next_observation)

      # The rewards and discounts have to have the same type as network values.
      r_t = tf.cast(transitions.reward, q_tm1.dtype)
      r_t = tf.clip_by_value(r_t, -1., 1.)
      d_t = tf.cast(transitions.discount, q_tm1.dtype) * tf.cast(
          self._discount, q_tm1.dtype)

      # Compute the loss.
      _, extra = trfl.double_qlearning(q_tm1, transitions.action, r_t, d_t,
                                       q_t_value, q_t_selector)
      loss = losses.huber(extra.td_error, self._huber_loss_parameter)

      # Get the importance weights.
      importance_weights = 1. / probs  # [B]
      importance_weights **= self._importance_sampling_exponent
      importance_weights /= tf.reduce_max(importance_weights)

      # Reweight.
      loss *= tf.cast(importance_weights, loss.dtype)  # [B]
      loss = tf.reduce_mean(loss, axis=[0])  # []

    # Do a step of SGD.
    gradients = tape.gradient(loss, self._network.trainable_variables)
    self._optimizer.apply(gradients, self._network.trainable_variables)

    # Update the priorities in the replay buffer.
    if self._replay_client:
      priorities = tf.cast(tf.abs(extra.td_error), tf.float64)
      self._replay_client.update_priorities(
          table=adders.DEFAULT_PRIORITY_TABLE, keys=keys, priorities=priorities)

    # Periodically update the target network.
    if tf.math.mod(self._num_steps, self._target_update_period) == 0:
      for src, dest in zip(self._q_network.variables,
                           self._target_q_network.variables):
        dest.assign(src)
    self._num_steps.assign_add(1)

    # Compute the global norm of the gradients for logging.
    global_gradient_norm = tf.linalg.global_norm(gradients)

    # Compute statistics of the Q-values for logging.
    max_q = tf.reduce_max(q_t_value)
    min_q = tf.reduce_min(q_t_value)
    mean_q, var_q = tf.nn.moments(q_t_value, [0, 1])

    # Report loss & statistics for logging.
    fetches = {
        'gradient_norm': global_gradient_norm,
        'loss': loss,
        'max_q': max_q,
        'mean_q': mean_q,
        'min_q': min_q,
        'var_q': var_q,
    }

    return fetches

  def step(self):
    # Do a batch of SGD.
    result = self._step()

    # Update our counts and record it.
    counts = self._counter.increment(steps=1)
    result.update(counts)

    # Snapshot and attempt to write logs.
    if self._snapshotter is not None:
      self._snapshotter.save()
    self._logger.write(result)

  def get_variables(self, names: List[str]) -> List[np.ndarray]:
    return tf2_utils.to_numpy(self._variables)

  @property
  def state(self):
    """Returns the stateful parts of the learner for checkpointing."""
    return {
        'network': self._network,
        'target_q_network': self._target_q_network,
        'optimizer': self._optimizer,
        'num_steps': self._num_steps
    }


class DiscreteBCQLearner(core.Learner, tf2_savers.TFSaveable):
  """Discrete BCQ learner.

  This learner combines supervised BC learning and Q learning to implement the
  discrete BCQ algorithm as described in https://arxiv.org/pdf/1910.01708.pdf.
  """

  def __init__(self,
               network: discrete_networks.DiscreteFilteredQNetwork,
               dataset: tf.data.Dataset,
               learning_rate: float,
               counter: Optional[counting.Counter] = None,
               bc_logger: Optional[loggers.Logger] = None,
               bcq_logger: Optional[loggers.Logger] = None,
               **bcq_learner_kwargs):
    counter = counter or counting.Counter()
    self._bc_logger = bc_logger or loggers.TerminalLogger('bc_learner',
                                                          time_delta=1.)
    self._bcq_logger = bcq_logger or loggers.TerminalLogger('bcq_learner',
                                                            time_delta=1.)

    self._bc_learner = bc.BCLearner(
        network=network.g_network,
        learning_rate=learning_rate,
        dataset=dataset,
        counter=counting.Counter(counter, 'bc'),
        logger=self._bc_logger,
        checkpoint=False)
    self._bcq_learner = _InternalBCQLearner(
        network=network,
        learning_rate=learning_rate,
        dataset=dataset,
        counter=counting.Counter(counter, 'bcq'),
        logger=self._bcq_logger,
        **bcq_learner_kwargs)

  def get_variables(self, names):
    return self._bcq_learner.get_variables(names)

  @property
  def state(self):
    bc_state = self._bc_learner.state
    bc_state.pop('network')  # No need to checkpoint the BC network.
    bcq_state = self._bcq_learner.state
    state = dict()
    state.update({f'bc_{k}': v for k, v in bc_state.items()})
    state.update({f'bcq_{k}': v for k, v in bcq_state.items()})
    return state

  def step(self):
    self._bc_learner.step()
    self._bcq_learner.step()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the ValueDice agent."""

from absl.testing import absltest
import acme
from acme import specs
from acme.agents.jax import value_dice
from acme.testing import fakes
from acme.utils import counting


class ValueDiceTest(absltest.TestCase):

  def test_value_dice(self):
    # Create a fake environment to test with.
    environment = fakes.ContinuousEnvironment(
        episode_length=10, action_dim=3, observation_dim=5, bounded=True)

    spec = specs.make_environment_spec(environment)

    # Create the networks.
    network = value_dice.make_networks(spec)

    config = value_dice.ValueDiceConfig(batch_size=10, min_replay_size=1)
    counter = counting.Counter()
    agent = value_dice.ValueDice(
        spec=spec,
        network=network,
        config=config,
        make_demonstrations=fakes.transition_iterator(environment),
        seed=0,
        counter=counter)

    # Try running the environment loop. We have no assertions here because all
    # we care about is that the agent runs without raising any errors.
    loop = acme.EnvironmentLoop(environment, agent, counter=counter)
    loop.run(num_episodes=2)


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DQN losses."""
import dataclasses
from typing import Tuple

from acme import types
from acme.agents.jax.dqn import learning_lib
from acme.jax import networks as networks_lib
import jax
import jax.numpy as jnp
import reverb
import rlax


@dataclasses.dataclass
class PrioritizedDoubleQLearning(learning_lib.LossFn):
  """Clipped double q learning with prioritization on TD error."""
  discount: float = 0.99
  importance_sampling_exponent: float = 0.2
  max_abs_reward: float = 1.
  huber_loss_parameter: float = 1.

  def __call__(
      self,
      network: networks_lib.FeedForwardNetwork,
      params: networks_lib.Params,
      target_params: networks_lib.Params,
      batch: reverb.ReplaySample,
      key: networks_lib.PRNGKey,
  ) -> Tuple[jnp.DeviceArray, learning_lib.LossExtra]:
    """Calculate a loss on a single batch of data."""
    del key
    transitions: types.Transition = batch.data
    keys, probs, *_ = batch.info

    # Forward pass.
    q_tm1 = network.apply(params, transitions.observation)
    q_t_value = network.apply(target_params, transitions.next_observation)
    q_t_selector = network.apply(params, transitions.next_observation)

    # Cast and clip rewards.
    d_t = (transitions.discount * self.discount).astype(jnp.float32)
    r_t = jnp.clip(transitions.reward, -self.max_abs_reward,
                   self.max_abs_reward).astype(jnp.float32)

    # Compute double Q-learning n-step TD-error.
    batch_error = jax.vmap(rlax.double_q_learning)
    td_error = batch_error(q_tm1, transitions.action, r_t, d_t, q_t_value,
                           q_t_selector)
    batch_loss = rlax.huber_loss(td_error, self.huber_loss_parameter)

    # Importance weighting.
    importance_weights = (1. / probs).astype(jnp.float32)
    importance_weights **= self.importance_sampling_exponent
    importance_weights /= jnp.max(importance_weights)

    # Reweight.
    loss = jnp.mean(importance_weights * batch_loss)  # []
    reverb_update = learning_lib.ReverbUpdate(
        keys=keys, priorities=jnp.abs(td_error).astype(jnp.float64))
    extra = learning_lib.LossExtra(metrics={}, reverb_update=reverb_update)
    return loss, extra


@dataclasses.dataclass
class PrioritizedCategoricalDoubleQLearning(learning_lib.LossFn):
  """Categorical double q learning with prioritization on TD error."""
  discount: float = 0.99
  importance_sampling_exponent: float = 0.2
  max_abs_reward: float = 1.

  def __call__(
      self,
      network: networks_lib.FeedForwardNetwork,
      params: networks_lib.Params,
      target_params: networks_lib.Params,
      batch: reverb.ReplaySample,
      key: networks_lib.PRNGKey,
  ) -> Tuple[jnp.DeviceArray, learning_lib.LossExtra]:
    """Calculate a loss on a single batch of data."""
    del key
    transitions: types.Transition = batch.data
    keys, probs, *_ = batch.info

    # Forward pass.
    _, logits_tm1, atoms_tm1 = network.apply(params, transitions.observation)
    _, logits_t, atoms_t = network.apply(target_params,
                                         transitions.next_observation)
    q_t_selector, _, _ = network.apply(params, transitions.next_observation)

    # Cast and clip rewards.
    d_t = (transitions.discount * self.discount).astype(jnp.float32)
    r_t = jnp.clip(transitions.reward, -self.max_abs_reward,
                   self.max_abs_reward).astype(jnp.float32)

    # Compute categorical double Q-learning loss.
    batch_loss_fn = jax.vmap(
        rlax.categorical_double_q_learning,
        in_axes=(None, 0, 0, 0, 0, None, 0, 0))
    batch_loss = batch_loss_fn(atoms_tm1, logits_tm1, transitions.action, r_t,
                               d_t, atoms_t, logits_t, q_t_selector)

    # Importance weighting.
    importance_weights = (1. / probs).astype(jnp.float32)
    importance_weights **= self.importance_sampling_exponent
    importance_weights /= jnp.max(importance_weights)

    # Reweight.
    loss = jnp.mean(importance_weights * batch_loss)  # []
    reverb_update = learning_lib.ReverbUpdate(
        keys=keys, priorities=jnp.abs(batch_loss).astype(jnp.float64))
    extra = learning_lib.LossExtra(metrics={}, reverb_update=reverb_update)
    return loss, extra


@dataclasses.dataclass
class QLearning(learning_lib.LossFn):
  """Clipped q learning.

  This matches the original DQN loss: https://arxiv.org/abs/1312.5602.
  """
  discount: float = 0.99
  max_abs_reward: float = 1.
  huber_loss_parameter: float = 1.

  def __call__(
      self,
      network: networks_lib.FeedForwardNetwork,
      params: networks_lib.Params,
      target_params: networks_lib.Params,
      batch: reverb.ReplaySample,
      key: networks_lib.PRNGKey,
  ) -> Tuple[jnp.DeviceArray, learning_lib.LossExtra]:
    """Calculate a loss on a single batch of data."""
    del key
    transitions: types.Transition = batch.data

    # Forward pass.
    q_tm1 = network.apply(params, transitions.observation)
    q_t = network.apply(target_params, transitions.next_observation)

    # Cast and clip rewards.
    d_t = (transitions.discount * self.discount).astype(jnp.float32)
    r_t = jnp.clip(transitions.reward, -self.max_abs_reward,
                   self.max_abs_reward).astype(jnp.float32)

    # Compute Q-learning TD-error.
    batch_error = jax.vmap(rlax.q_learning)
    td_error = batch_error(q_tm1, transitions.action, r_t, d_t, q_t)
    batch_loss = rlax.huber_loss(td_error, self.huber_loss_parameter)

    loss = jnp.mean(batch_loss)
    extra = learning_lib.LossExtra(metrics={})
    return loss, extra


@dataclasses.dataclass
class RegularizedQLearning(learning_lib.LossFn):
  """Regularized Q-learning.

  Implements DQNReg loss function: https://arxiv.org/abs/2101.03958.
  This is almost identical to QLearning except: 1) Adds a regularization term;
  2) Uses vanilla TD error without huber loss. 3) No reward clipping.
  """
  discount: float = 0.99
  regularizer_coeff = 0.1

  def __call__(
      self,
      network: networks_lib.FeedForwardNetwork,
      params: networks_lib.Params,
      target_params: networks_lib.Params,
      batch: reverb.ReplaySample,
      key: networks_lib.PRNGKey,
  ) -> Tuple[jnp.DeviceArray, learning_lib.LossExtra]:
    """Calculate a loss on a single batch of data."""
    del key
    transitions: types.Transition = batch.data

    # Forward pass.
    q_tm1 = network.apply(params, transitions.observation)
    q_t = network.apply(target_params, transitions.next_observation)

    d_t = (transitions.discount * self.discount).astype(jnp.float32)

    # Compute Q-learning TD-error.
    batch_error = jax.vmap(rlax.q_learning)
    td_error = batch_error(
        q_tm1, transitions.action, transitions.reward, d_t, q_t)
    td_error = 0.5 * jnp.square(td_error)

    def select(qtm1, action):
      return qtm1[action]
    q_regularizer = jax.vmap(select)(q_tm1, transitions.action)

    loss = self.regularizer_coeff * jnp.mean(q_regularizer) + jnp.mean(td_error)
    extra = learning_lib.LossExtra(metrics={})
    return loss, extra


@dataclasses.dataclass
class MunchausenQLearning(learning_lib.LossFn):
  """Munchausen q learning.

  Implements M-DQN: https://arxiv.org/abs/2007.14430.
  """
  entropy_temperature: float = 0.03  # tau parameter
  munchausen_coefficient: float = 0.9  # alpha parameter
  clip_value_min: float = -1e3
  discount: float = 0.99
  max_abs_reward: float = 1.
  huber_loss_parameter: float = 1.

  def __call__(
      self,
      network: networks_lib.FeedForwardNetwork,
      params: networks_lib.Params,
      target_params: networks_lib.Params,
      batch: reverb.ReplaySample,
      key: networks_lib.PRNGKey,
  ) -> Tuple[jnp.DeviceArray, learning_lib.LossExtra]:
    """Calculate a loss on a single batch of data."""
    del key
    transitions: types.Transition = batch.data

    # Forward pass.
    q_online_s = network.apply(params, transitions.observation)
    action_one_hot = jax.nn.one_hot(transitions.action, q_online_s.shape[-1])
    q_online_sa = jnp.sum(action_one_hot * q_online_s, axis=-1)
    q_target_s = network.apply(target_params, transitions.observation)
    q_target_next = network.apply(target_params, transitions.next_observation)

    # Cast and clip rewards.
    d_t = (transitions.discount * self.discount).astype(jnp.float32)
    r_t = jnp.clip(transitions.reward, -self.max_abs_reward,
                   self.max_abs_reward).astype(jnp.float32)

    # Munchausen term : tau * log_pi(a|s)
    munchausen_term = self.entropy_temperature * jax.nn.log_softmax(
        q_target_s / self.entropy_temperature, axis=-1)
    munchausen_term_a = jnp.sum(action_one_hot * munchausen_term, axis=-1)
    munchausen_term_a = jnp.clip(munchausen_term_a,
                                 a_min=self.clip_value_min,
                                 a_max=0.)

    # Soft Bellman operator applied to q
    next_v = self.entropy_temperature * jax.nn.logsumexp(
        q_target_next / self.entropy_temperature, axis=-1)
    target_q = jax.lax.stop_gradient(r_t + self.munchausen_coefficient *
                                     munchausen_term_a + d_t * next_v)

    batch_loss = rlax.huber_loss(target_q - q_online_sa,
                                 self.huber_loss_parameter)
    loss = jnp.mean(batch_loss)

    extra = learning_lib.LossExtra(metrics={})
    return loss, extra



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for PPO agent."""

from typing import Callable, Tuple

from absl.testing import absltest
import acme
from acme import specs
from acme.agents.jax import ppo
from acme.jax import networks as networks_lib
from acme.jax import utils
from acme.testing import fakes
from acme.utils import counting
from acme.utils import loggers
import flax
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
import tensorflow_probability
tfd = tensorflow_probability.substrates.jax.distributions


def make_haiku_networks(
    spec: specs.EnvironmentSpec) -> networks_lib.FeedForwardNetwork:
  """Creates Haiku networks to be used by the agent."""

  num_actions = spec.actions.num_values

  def forward_fn(inputs):
    policy_network = hk.Sequential([
        utils.batch_concat,
        hk.nets.MLP([64, 64]),
        networks_lib.CategoricalHead(num_actions)
    ])
    value_network = hk.Sequential([
        utils.batch_concat,
        hk.nets.MLP([64, 64]),
        hk.Linear(1), lambda x: jnp.squeeze(x, axis=-1)
    ])

    action_distribution = policy_network(inputs)
    value = value_network(inputs)
    return (action_distribution, value)

  # Transform into pure functions.
  forward_fn = hk.without_apply_rng(hk.transform(forward_fn))

  dummy_obs = utils.zeros_like(spec.observations)
  dummy_obs = utils.add_batch_dim(dummy_obs)  # Dummy 'sequence' dim.
  return networks_lib.FeedForwardNetwork(
      lambda rng: forward_fn.init(rng, dummy_obs), forward_fn.apply)


def make_flax_networks(
    spec: specs.EnvironmentSpec) -> networks_lib.FeedForwardNetwork:
  """Creates FLAX networks to be used by the agent."""

  num_actions = spec.actions.num_values

  class MLP(flax.nn.Module):
    """MLP module."""

    def apply(self,
              data: jnp.ndarray,
              layer_sizes: Tuple[int],
              activation: Callable[[jnp.ndarray], jnp.ndarray] = flax.nn.relu,
              kernel_init: object = jax.nn.initializers.lecun_uniform(),
              activate_final: bool = False,
              bias: bool = True):
      hidden = data
      for i, hidden_size in enumerate(layer_sizes):
        hidden = flax.nn.Dense(
            hidden,
            hidden_size,
            name=f'hidden_{i}',
            kernel_init=kernel_init,
            bias=bias)
        if i != len(layer_sizes) - 1 or activate_final:
          hidden = activation(hidden)
      return hidden

  class PolicyValueModule(flax.nn.Module):
    """MLP module."""

    def apply(self, inputs: jnp.ndarray):
      inputs = utils.batch_concat(inputs)
      logits = MLP(inputs, [64, 64, num_actions])
      value = MLP(inputs, [64, 64, 1])
      value = jnp.squeeze(value, axis=-1)
      return tfd.Categorical(logits=logits), value

  dummy_obs = utils.zeros_like(spec.observations)
  dummy_obs = utils.add_batch_dim(dummy_obs)  # Dummy 'sequence' dim.
  return networks_lib.FeedForwardNetwork(
      lambda rng: PolicyValueModule.init(rng, dummy_obs)[1],
      PolicyValueModule.call)


class PPOTest(absltest.TestCase):

  def test_ppo_with_haiku(self):
    self.run_ppo_agent(make_haiku_networks)

  def test_ppo_with_flax(self):
    self.run_ppo_agent(make_flax_networks)

  def run_ppo_agent(self, make_networks_fn):
    # Create a fake environment to test with.
    environment = fakes.DiscreteEnvironment(
        num_actions=5,
        num_observations=10,
        obs_shape=(10, 5),
        obs_dtype=np.float32,
        episode_length=10)
    spec = specs.make_environment_spec(environment)

    distribution_value_networks = make_networks_fn(spec)
    ppo_networks = ppo.make_ppo_networks(distribution_value_networks)
    config = ppo.PPOConfig(unroll_length=4, num_epochs=2, num_minibatches=2)
    workdir = self.create_tempdir()
    counter = counting.Counter()
    logger_fn = lambda: loggers.make_default_logger('learner')
    # Construct the agent.
    agent = ppo.PPO(
        spec=spec,
        networks=ppo_networks,
        config=config,
        seed=0,
        workdir=workdir.full_path,
        normalize_input=True,
        counter=counter,
        logger_fn=logger_fn,
    )

    # Try running the environment loop. We have no assertions here because all
    # we care about is that the agent runs without raising any errors.
    loop = acme.EnvironmentLoop(environment, agent, counter=counter)
    loop.run(num_episodes=20)

  def test_ppo_nest_safety(self):
    # Create a fake environment with nested observations.
    environment = fakes.NestedDiscreteEnvironment(
        num_observations={
            'lat': 2,
            'long': 3
        },
        num_actions=5,
        obs_shape=(10, 5),
        obs_dtype=np.float32,
        episode_length=10)
    spec = specs.make_environment_spec(environment)

    distribution_value_networks = make_haiku_networks(spec)
    ppo_networks = ppo.make_ppo_networks(distribution_value_networks)
    config = ppo.PPOConfig(unroll_length=4, num_epochs=2, num_minibatches=2)
    workdir = self.create_tempdir()
    # Construct the agent.
    agent = ppo.PPO(
        spec=spec,
        networks=ppo_networks,
        config=config,
        seed=0,
        workdir=workdir.full_path,
        normalize_input=True,
    )

    # Try running the environment loop. We have no assertions here because all
    # we care about is that the agent runs without raising any errors.
    loop = acme.EnvironmentLoop(environment, agent)
    loop.run(num_episodes=20)


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines distributed and local ValueDice agents, using JAX."""

import functools
from typing import Callable, Iterator, Optional

from acme import specs
from acme import types
from acme.agents.jax.value_dice import builder
from acme.agents.jax.value_dice import config as value_dice_config
from acme.agents.jax.value_dice import networks
from acme.jax import utils
from acme.jax.layouts import distributed_layout
from acme.jax.layouts import local_layout
from acme.utils import counting
from acme.utils import loggers
import dm_env


NetworkFactory = Callable[[specs.EnvironmentSpec], networks.ValueDiceNetworks]


class DistributedValueDice(distributed_layout.DistributedLayout):
  """Distributed program definition for ValueDice.
  """

  def __init__(
      self,
      environment_factory: Callable[[bool], dm_env.Environment],
      network_factory: NetworkFactory,
      config: value_dice_config.ValueDiceConfig,
      make_demonstrations: Callable[[int], Iterator[types.Transition]],
      seed: int,
      num_actors: int,
      max_number_of_steps: Optional[int] = None,
      log_to_bigtable: bool = False,
      log_every: float = 10.0,
  ):
    logger_fn = functools.partial(loggers.make_default_logger,
                                  'learner', log_to_bigtable,
                                  time_delta=log_every, asynchronous=True,
                                  serialize_fn=utils.fetch_devicearray,
                                  steps_key='learner_steps')
    spec = specs.make_environment_spec(environment_factory(False))
    value_dice_builder = builder.ValueDiceBuilder(
        config=config, logger_fn=logger_fn,
        make_demonstrations=make_demonstrations)
    eval_policy_factory = (
        lambda n: networks.apply_policy_and_sample(n, True))
    super().__init__(
        seed=seed,
        environment_spec=spec,
        environment_factory=lambda: environment_factory(False),
        network_factory=network_factory,
        builder=value_dice_builder,
        policy_network=networks.apply_policy_and_sample,
        evaluator_factories=[
            distributed_layout.default_evaluator(
                environment_factory=lambda: environment_factory(True),
                network_factory=network_factory,
                builder=value_dice_builder,
                policy_factory=eval_policy_factory,
                log_to_bigtable=log_to_bigtable)
        ],
        num_actors=num_actors,
        max_number_of_steps=max_number_of_steps,
        prefetch_size=config.prefetch_size,
        log_to_bigtable=log_to_bigtable,
        actor_logger_fn=distributed_layout.get_default_logger_fn(
            log_to_bigtable, log_every),
    )


class ValueDice(local_layout.LocalLayout):
  """Local agent for ValueDice.
  """

  def __init__(
      self,
      spec: specs.EnvironmentSpec,
      network: networks.ValueDiceNetworks,
      config: value_dice_config.ValueDiceConfig,
      make_demonstrations: Callable[[int], Iterator[types.Transition]],
      seed: int,
      counter: Optional[counting.Counter] = None,
  ):
    min_replay_size = config.min_replay_size
    # Local layout (actually agent.Agent) makes sure that we populate the
    # buffer with min_replay_size initial transitions and that there's no need
    # for tolerance_rate. In order for deadlocks not to happen we need to
    # disable rate limiting that heppens inside the ValueDiceBuilder.
    # This is achieved by the following two lines.
    config.samples_per_insert_tolerance_rate = float('inf')
    config.min_replay_size = 1
    self.builder = builder.ValueDiceBuilder(
        config=config, make_demonstrations=make_demonstrations)
    super().__init__(
        seed=seed,
        environment_spec=spec,
        builder=self.builder,
        networks=network,
        policy_network=networks.apply_policy_and_sample(network),
        batch_size=config.batch_size,
        samples_per_insert=config.samples_per_insert,
        min_replay_size=min_replay_size,
        num_sgd_steps_per_step=config.num_sgd_steps_per_step,
        counter=counter,
    )



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""R2D2 actor."""

from typing import Generic, Mapping

from acme.agents.jax import actor_core as actor_core_lib
from acme.agents.jax.r2d2 import networks as r2d2_networks
from acme.jax import networks as networks_lib
from acme.jax import utils
import chex
import jax
import jax.numpy as jnp


@chex.dataclass(frozen=True, mappable_dataclass=False)
class R2D2ActorState(Generic[actor_core_lib.RecurrentState]):
  rng: networks_lib.PRNGKey
  epsilon: jnp.ndarray
  recurrent_state: actor_core_lib.RecurrentState


def get_actor_core(
    recurrent_policy: r2d2_networks.EpsilonRecurrentPolicy[
        actor_core_lib.RecurrentState],
    initial_core_state: actor_core_lib.RecurrentState, num_epsilons: int
) -> actor_core_lib.ActorCore[R2D2ActorState[actor_core_lib.RecurrentState],
                              Mapping[str, jnp.ndarray]]:
  """Returns ActorCore for R2D2."""
  def select_action(params: networks_lib.Params,
                    observation: networks_lib.Observation,
                    state: R2D2ActorState[actor_core_lib.RecurrentState]):
    # TODO(b/161332815): Make JAX Actor work with batched or unbatched inputs.
    rng, policy_rng = jax.random.split(state.rng)
    observation = utils.add_batch_dim(observation)
    recurrent_state = utils.add_batch_dim(state.recurrent_state)
    action, new_recurrent_state = utils.squeeze_batch_dim(recurrent_policy(
        params, policy_rng, observation, recurrent_state, state.epsilon))
    return action, R2D2ActorState(rng, state.epsilon, new_recurrent_state)

  initial_core_state = utils.squeeze_batch_dim(initial_core_state)

  def init(
      rng: networks_lib.PRNGKey
  ) -> R2D2ActorState[actor_core_lib.RecurrentState]:
    rng, epsilon_rng = jax.random.split(rng)
    epsilon = jax.random.choice(epsilon_rng,
                                jnp.logspace(1, 8, num_epsilons, base=0.4))
    return R2D2ActorState(rng, epsilon, initial_core_state)

  def get_extras(
      state: R2D2ActorState[actor_core_lib.RecurrentState]
  ) -> Mapping[str, jnp.ndarray]:
    return {'core_state': state.recurrent_state}

  return actor_core_lib.ActorCore(init=init, select_action=select_action,
                                  get_extras=get_extras)



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common tools for reverb replay."""

import dataclasses
from typing import Any, Callable, Dict, Iterator, Optional

from acme import adders as adders_lib
from acme import datasets
from acme import specs
from acme import types
from acme.adders import reverb as adders
import reverb


@dataclasses.dataclass
class ReverbReplay:
  server: reverb.Server
  adder: adders_lib.Adder
  data_iterator: Iterator[reverb.ReplaySample]
  client: Optional[reverb.Client] = None
  can_sample: Callable[[], bool] = lambda: True


def make_reverb_prioritized_nstep_replay(
    environment_spec: specs.EnvironmentSpec,
    extra_spec: types.NestedSpec = (),
    n_step: int = 1,
    batch_size: int = 32,
    max_replay_size: int = 100_000,
    min_replay_size: int = 1,
    discount: float = 1.,
    prefetch_size: int = 4,  # TODO(iosband): rationalize prefetch size.
    replay_table_name: str = adders.DEFAULT_PRIORITY_TABLE,
    priority_exponent: Optional[float] = None,  # If None, default to uniform.
) -> ReverbReplay:
  """Creates a single-process replay infrastructure from an environment spec."""
  # Parsing priority exponent to determine uniform vs prioritized replay
  if priority_exponent is None:
    sampler = reverb.selectors.Uniform()
    priority_fns = {replay_table_name: lambda x: 1.}
  else:
    sampler = reverb.selectors.Prioritized(priority_exponent)
    priority_fns = None

  # Create a replay server to add data to. This uses no limiter behavior in
  # order to allow the Agent interface to handle it.
  replay_table = reverb.Table(
      name=replay_table_name,
      sampler=sampler,
      remover=reverb.selectors.Fifo(),
      max_size=max_replay_size,
      rate_limiter=reverb.rate_limiters.MinSize(min_replay_size),
      signature=adders.NStepTransitionAdder.signature(environment_spec,
                                                        extra_spec),
  )
  server = reverb.Server([replay_table], port=None)

  # The adder is used to insert observations into replay.
  address = f'localhost:{server.port}'
  client = reverb.Client(address)
  adder = adders.NStepTransitionAdder(
      client, n_step, discount, priority_fns=priority_fns)

  # The dataset provides an interface to sample from replay.
  data_iterator = datasets.make_reverb_dataset(
      table=replay_table_name,
      server_address=address,
      batch_size=batch_size,
      prefetch_size=prefetch_size,
  ).as_numpy_iterator()
  return ReverbReplay(server, adder, data_iterator, client=client)


def make_reverb_online_queue(
    environment_spec: specs.EnvironmentSpec,
    extra_spec: Dict[str, Any],
    max_queue_size: int,
    sequence_length: int,
    sequence_period: int,
    batch_size: int,
    replay_table_name: str = adders.DEFAULT_PRIORITY_TABLE,
) -> ReverbReplay:
  """Creates a single process queue from an environment spec and extra_spec."""
  signature = adders.SequenceAdder.signature(environment_spec, extra_spec)
  queue = reverb.Table.queue(
      name=replay_table_name, max_size=max_queue_size, signature=signature)
  server = reverb.Server([queue], port=None)
  can_sample = lambda: queue.can_sample(batch_size)

  # Component to add things into replay.
  address = f'localhost:{server.port}'
  adder = adders.SequenceAdder(
      client=reverb.Client(address),
      period=sequence_period,
      sequence_length=sequence_length,
  )

  # The dataset object to learn from.
  # We don't use datasets.make_reverb_dataset() here to avoid interleaving
  # and prefetching, that doesn't work well with can_sample() check on update.
  dataset = reverb.TrajectoryDataset.from_table_signature(
      server_address=address,
      table=replay_table_name,
      max_in_flight_samples_per_worker=1,
  )
  dataset = dataset.batch(batch_size, drop_remainder=True)
  data_iterator = dataset.as_numpy_iterator()
  return ReverbReplay(server, adder, data_iterator, can_sample=can_sample)


def make_reverb_prioritized_sequence_replay(
    environment_spec: specs.EnvironmentSpec,
    extra_spec: types.NestedSpec = (),
    batch_size: int = 32,
    max_replay_size: int = 100_000,
    min_replay_size: int = 1,
    priority_exponent: float = 0.,
    burn_in_length: int = 40,
    sequence_length: int = 80,
    sequence_period: int = 40,
    replay_table_name: str = adders.DEFAULT_PRIORITY_TABLE,
    prefetch_size: int = 4,
) -> ReverbReplay:
  """Single-process replay for sequence data from an environment spec."""
  # Create a replay server to add data to. This uses no limiter behavior in
  # order to allow the Agent interface to handle it.
  replay_table = reverb.Table(
      name=replay_table_name,
      sampler=reverb.selectors.Prioritized(priority_exponent),
      remover=reverb.selectors.Fifo(),
      max_size=max_replay_size,
      rate_limiter=reverb.rate_limiters.MinSize(min_replay_size),
      signature=adders.SequenceAdder.signature(environment_spec, extra_spec),
  )
  server = reverb.Server([replay_table], port=None)

  # The adder is used to insert observations into replay.
  address = f'localhost:{server.port}'
  client = reverb.Client(address)
  sequence_length = burn_in_length + sequence_length + 1
  adder = adders.SequenceAdder(
      client=client,
      period=sequence_period,
      sequence_length=sequence_length,
      delta_encoded=True,
  )

  # The dataset provides an interface to sample from replay.
  data_iterator = datasets.make_reverb_dataset(
      table=replay_table_name,
      server_address=address,
      batch_size=batch_size,
      prefetch_size=prefetch_size,
  ).as_numpy_iterator()
  return ReverbReplay(server, adder, data_iterator, client)



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines distributed and local TD3 agents, using JAX."""

import functools
from typing import Callable, Optional

from acme import specs
from acme.agents.jax.td3 import builder
from acme.agents.jax.td3 import config as td3_config
from acme.agents.jax.td3 import networks
from acme.jax import utils
from acme.jax.layouts import distributed_layout
from acme.jax.layouts import local_layout
from acme.utils import counting
from acme.utils import loggers
import dm_env


NetworkFactory = Callable[[specs.EnvironmentSpec], networks.TD3Networks]


class DistributedTD3(distributed_layout.DistributedLayout):
  """Distributed program definition for TD3.
  """

  def __init__(
      self,
      environment_factory: Callable[[bool], dm_env.Environment],
      environment_spec: specs.EnvironmentSpec,
      network_factory: NetworkFactory,
      config: td3_config.TD3Config,
      seed: int,
      num_actors: int,
      max_number_of_steps: Optional[int] = None,
      log_to_bigtable: bool = False,
      log_every: float = 10.0,
  ):
    logger_fn = functools.partial(loggers.make_default_logger,
                                  'learner', log_to_bigtable,
                                  time_delta=log_every, asynchronous=True,
                                  serialize_fn=utils.fetch_devicearray,
                                  steps_key='learner_steps')
    td3_builder = builder.TD3Builder(config, logger_fn=logger_fn)

    action_specs = environment_spec.actions
    policy_network_fn = functools.partial(networks.get_default_behavior_policy,
                                          action_specs=action_specs,
                                          sigma=config.sigma)

    eval_network_fn = functools.partial(networks.get_default_behavior_policy,
                                        action_specs=action_specs,
                                        sigma=0.)
    super().__init__(
        seed=seed,
        environment_factory=lambda: environment_factory(False),
        network_factory=network_factory,
        builder=td3_builder,
        policy_network=policy_network_fn,
        evaluator_factories=[
            distributed_layout.default_evaluator(
                environment_factory=lambda: environment_factory(True),
                network_factory=network_factory,
                builder=td3_builder,
                policy_factory=eval_network_fn,
                log_to_bigtable=log_to_bigtable)
        ],
        num_actors=num_actors,
        max_number_of_steps=max_number_of_steps,
        prefetch_size=config.prefetch_size,
        log_to_bigtable=log_to_bigtable,
        actor_logger_fn=distributed_layout.get_default_logger_fn(
            log_to_bigtable, log_every),
    )


class TD3(local_layout.LocalLayout):
  """Local agent for TD3.
  """

  def __init__(
      self,
      spec: specs.EnvironmentSpec,
      network: networks.TD3Networks,
      config: td3_config.TD3Config,
      seed: int,
      counter: Optional[counting.Counter] = None,
  ):
    min_replay_size = config.min_replay_size
    # Local layout (actually agent.Agent) makes sure that we populate the
    # buffer with min_replay_size initial transitions and that there's no need
    # for tolerance_rate. In order to avoid deadlocks, we disable rate limiting
    # that is configured in TD3Builder.make_replay_tables. This is achieved by
    # the following two lines.
    config.samples_per_insert_tolerance_rate = float('inf')
    config.min_replay_size = 1

    behavior_policy = networks.get_default_behavior_policy(
        networks=network,
        action_specs=spec.actions,
        sigma=config.sigma)

    self.builder = builder.TD3Builder(config)
    super().__init__(
        seed=seed,
        environment_spec=spec,
        builder=self.builder,
        networks=network,
        policy_network=behavior_policy,
        batch_size=config.batch_size,
        prefetch_size=config.prefetch_size,
        samples_per_insert=config.samples_per_insert,
        min_replay_size=min_replay_size,
        num_sgd_steps_per_step=config.num_sgd_steps_per_step,
        counter=counter,
    )



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Variable utilities for JAX."""

from concurrent import futures
from typing import List, Optional, Sequence, Union

from acme import core
from acme.jax import networks as network_types

import jax


class VariableClient:
  """A variable client for updating variables from a remote source."""

  def __init__(self,
               client: core.VariableSource,
               key: Union[str, Sequence[str]],
               update_period: int = 1,
               device: Optional[Union[str, jax.xla.Device]] = None):
    """Initializes the variable client.

    Args:
      client: A variable source from which we fetch variables.
      key: Which variables to request. When multiple keys are used, params
        property will return a list of params.
      update_period: Interval between fetches.
      device: The name of a JAX device to put variables on. If None (default),
        don't put to device.
    """
    self._update_period = update_period
    self._call_counter = 0
    self._client = client
    self._params: Sequence[network_types.Params] = None

    self._device = device
    if isinstance(self._device, str):
      self._device = jax.devices(device)[0]

    self._executor = futures.ThreadPoolExecutor(max_workers=1)

    if isinstance(key, str):
      key = [key]

    self._key = key
    self._request = lambda k=key: client.get_variables(k)
    self._future: Optional[futures.Future] = None
    self._async_request = lambda: self._executor.submit(self._request)

  def update(self, wait: bool = False) -> None:
    """Periodically updates the variables with the latest copy from the source.

    If wait is True, a blocking request is executed. Any active request will be
    cancelled.
    If wait is False, this method makes an asynchronous request for variables.

    Args:
      wait: Whether to execute asynchronous (False) or blocking updates (True).
        Defaults to False.
    """
    # Track calls (we only update periodically).
    if self._call_counter < self._update_period:
      self._call_counter += 1

    # Return if it's not time to fetch another update.
    if self._call_counter < self._update_period:
      return

    if wait:
      if self._future is not None:
        if self._future.running():
          self._future.cancel()
        self._future = None
      self._call_counter = 0
      self.update_and_wait()
      return

    # Return early if we are still waiting for a previous request to come back.
    if self._future and not self._future.done():
      return

    # Get a future and add the copy function as a callback.
    self._call_counter = 0
    self._future = self._async_request()
    self._future.add_done_callback(lambda f: self._callback(f.result()))

  def update_and_wait(self):
    """Immediately update and block until we get the result."""
    self._callback(self._request())

  def _callback(self, params_list: List[network_types.Params]):
    if self._device:
      # Move variables to a proper device.
      self._params = jax.device_put(params_list, self._device)
    else:
      self._params = params_list

  @property
  def device(self) -> Optional[jax.xla.Device]:
    return self._device

  @property
  def params(self) -> Union[network_types.Params, List[network_types.Params]]:
    """Returns the first params for one key, otherwise the whole params list."""
    if self._params is None:
      self.update_and_wait()

    if len(self._params) == 1:
      return self._params[0]
    else:
      return self._params



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PPO config."""
import dataclasses
from typing import Callable, Union

from acme.adders import reverb as adders_reverb
import numpy as np


@dataclasses.dataclass
class PPOConfig:
  """Configuration options for PPO."""
  unroll_length: int
  num_minibatches: int
  num_epochs: int

  batch_size: int = 1
  clip_value: bool = False
  replay_table_name: str = adders_reverb.DEFAULT_PRIORITY_TABLE
  ppo_clipping_epsilon: float = 0.2
  gae_lambda: float = 0.95
  discount: float = 0.99
  learning_rate: Union[float, Callable[[int], float]] = 1e-3
  adam_epsilon: float = 1e-5
  entropy_cost: float = 0.01
  value_cost: float = 1.
  max_abs_reward: float = np.inf
  max_gradient_norm: float = 0.5
  prefetch_size: int = 4
  variable_update_period: int = 1



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sequence adders.

This implements adders which add sequences or partial trajectories.
"""

import enum
import operator
from typing import Iterable, Optional

from acme import specs
from acme import types
from acme.adders.reverb import base
from acme.adders.reverb import utils

import numpy as np
import reverb
import tensorflow as tf
import tree


class EndBehavior(enum.Enum):
  """Class to enumerate available options for writing behavior at episode ends.

  Example:

    sequence_length = 3
    period = 2

  Episode steps (digits) and writing events (W):

             1 2 3 4 5 6
                 W   W

  First two sequences:

             1 2 3
             . . 3 4 5

  Written sequences for the different end of episode behaviors:
  Here are the last written sequences for each end of episode behavior:

   WRITE     . . . 4 5 6
   CONTINUE  . . . . 5 6 F
   ZERO_PAD  . . . . 5 6 0
   TRUNCATE  . . . . 5 6

  Key:
    F: First step of the next episode
    0: Zero-filled Step
  """
  WRITE = 'write_buffer'
  CONTINUE = 'continue_to_next_episode'
  ZERO_PAD = 'zero_pad_til_next_write'
  TRUNCATE = 'write_truncated_buffer'


class SequenceAdder(base.ReverbAdder):
  """An adder which adds sequences of fixed length."""

  def __init__(
      self,
      client: reverb.Client,
      sequence_length: int,
      period: int,
      *,
      delta_encoded: bool = False,
      priority_fns: Optional[base.PriorityFnMapping] = None,
      max_in_flight_items: Optional[int] = 2,
      end_of_episode_behavior: Optional[EndBehavior] = None,
      # Deprecated kwargs.
      chunk_length: Optional[int] = None,
      pad_end_of_episode: Optional[bool] = None,
      break_end_of_episode: Optional[bool] = None,
      get_signature_timeout_ms: int = 300_000,
  ):
    """Makes a SequenceAdder instance.

    Args:
      client: See docstring for BaseAdder.
      sequence_length: The fixed length of sequences we wish to add.
      period: The period with which we add sequences. If less than
        sequence_length, overlapping sequences are added. If equal to
        sequence_length, sequences are exactly non-overlapping.
      delta_encoded: If `True` (False by default) enables delta encoding, see
        `Client` for more information.
      priority_fns: See docstring for BaseAdder.
      max_in_flight_items: The maximum number of items allowed to be "in flight"
        at the same time. See `block_until_num_items` in
        `reverb.TrajectoryWriter.flush` for more info.
      end_of_episode_behavior:  Determines how sequences at the end of the
        episode are handled (default `EndOfEpisodeBehavior.ZERO_PAD`). See
        the docstring for `EndOfEpisodeBehavior` for more information.
      chunk_length: Deprecated and unused.
      pad_end_of_episode: If True (default) then upon end of episode the current
        sequence will be padded (with observations, actions, etc... whose values
        are 0) until its length is `sequence_length`. If False then the last
        sequence in the episode may have length less than `sequence_length`.
      break_end_of_episode: If 'False' (True by default) does not break
        sequences on env reset. In this case 'pad_end_of_episode' is not used.
      get_signature_timeout_ms: time before timeout in fetching the signature
        from the reverb server.
    """
    del chunk_length
    super().__init__(
        client=client,
        # We need an additional space in the buffer for the partial step the
        # base.ReverbAdder will add with the next observation.
        max_sequence_length=sequence_length+1,
        delta_encoded=delta_encoded,
        priority_fns=priority_fns,
        max_in_flight_items=max_in_flight_items,
        get_signature_timeout_ms=get_signature_timeout_ms)

    if pad_end_of_episode and not break_end_of_episode:
      raise ValueError(
          'Can\'t set pad_end_of_episode=True and break_end_of_episode=False at'
          ' the same time, since those behaviors are incompatible.')

    self._period = period
    self._sequence_length = sequence_length

    if end_of_episode_behavior and (pad_end_of_episode is not None or
                                    break_end_of_episode is not None):
      raise ValueError(
          'Using end_of_episode_behavior and either '
          'pad_end_of_episode or break_end_of_episode is not permitted. '
          'Please use only end_of_episode_behavior instead.')

    # Set pad_end_of_episode and break_end_of_episode to default values.
    if end_of_episode_behavior is None and pad_end_of_episode is None:
      pad_end_of_episode = True
    if end_of_episode_behavior is None and break_end_of_episode is None:
      break_end_of_episode = True

    self._end_of_episode_behavior = EndBehavior.ZERO_PAD
    if pad_end_of_episode is not None or break_end_of_episode is not None:
      if not break_end_of_episode:
        self._end_of_episode_behavior = EndBehavior.CONTINUE
      elif break_end_of_episode and pad_end_of_episode:
        self._end_of_episode_behavior = EndBehavior.ZERO_PAD
      elif break_end_of_episode and not pad_end_of_episode:
        self._end_of_episode_behavior = EndBehavior.TRUNCATE
      else:
        raise ValueError(
            'Reached an unexpected configuration of the SequenceAdder '
            f'with break_end_of_episode={break_end_of_episode} '
            f'and pad_end_of_episode={pad_end_of_episode}.')
    elif isinstance(end_of_episode_behavior, EndBehavior):
      self._end_of_episode_behavior = end_of_episode_behavior
    else:
      raise ValueError('end_of_episod_behavior must be an instance of '
                       f'EndBehavior, received {end_of_episode_behavior}.')

  def reset(self):
    """Resets the adder's buffer."""
    # If we do not write on end of episode, we should not reset the writer.
    if self._end_of_episode_behavior is EndBehavior.CONTINUE:
      return

    super().reset()

  def _write(self):
    self._maybe_create_item(self._sequence_length)

  def _write_last(self):
    # Maybe determine the delta to the next time we would write a sequence.
    if self._end_of_episode_behavior in (EndBehavior.TRUNCATE,
                                         EndBehavior.ZERO_PAD):
      delta = self._sequence_length - self._writer.episode_steps
      if delta < 0:
        delta = (self._period + delta) % self._period

    # Handle various end-of-episode cases.
    if self._end_of_episode_behavior is EndBehavior.CONTINUE:
      self._maybe_create_item(self._sequence_length, end_of_episode=True)

    elif self._end_of_episode_behavior is EndBehavior.WRITE:
      # Drop episodes that are too short.
      if self._writer.episode_steps < self._sequence_length:
        return
      self._maybe_create_item(
          self._sequence_length, end_of_episode=True, force=True)

    elif self._end_of_episode_behavior is EndBehavior.TRUNCATE:
      self._maybe_create_item(
          self._sequence_length - delta,
          end_of_episode=True,
          force=True)

    elif self._end_of_episode_behavior is EndBehavior.ZERO_PAD:
      zero_step = tree.map_structure(lambda x: np.zeros_like(x[-2].numpy()),
                                     self._writer.history)
      for _ in range(delta):
        self._writer.append(zero_step)

      self._maybe_create_item(
          self._sequence_length, end_of_episode=True, force=True)
    else:
      raise ValueError(
          f'Unhandled end of episode behavior: {self._end_of_episode_behavior}.'
          ' This should never happen, please contact Acme dev team.')

  def _maybe_create_item(self,
                         sequence_length: int,
                         *,
                         end_of_episode: bool = False,
                         force: bool = False):

    # Check conditions under which a new item is created.
    first_write = self._writer.episode_steps == sequence_length
    # NOTE(bshahr): the following line assumes that the only way sequence_length
    # is less than self._sequence_length, is if the episode is shorter than
    # self._sequence_length.
    period_reached = (
        self._writer.episode_steps > self._sequence_length and
        ((self._writer.episode_steps - self._sequence_length) % self._period
         == 0))

    if not first_write and not period_reached and not force:
      return

    # TODO(b/183945808): will need to change to adhere to the new protocol.
    if not end_of_episode:
      get_traj = operator.itemgetter(slice(-sequence_length - 1, -1))
    else:
      get_traj = operator.itemgetter(slice(-sequence_length, None))

    history = self._writer.history
    trajectory = base.Trajectory(**tree.map_structure(get_traj, history))

    # Compute priorities for the buffer.
    table_priorities = utils.calculate_priorities(self._priority_fns,
                                                  trajectory)

    # Create a prioritized item for each table.
    for table_name, priority in table_priorities.items():
      self._writer.create_item(table_name, priority, trajectory)
      self._writer.flush(self._max_in_flight_items)

  # TODO(bshahr): make this into a standalone method. Class methods should be
  # used as alternative constructors or when modifying some global state,
  # neither of which is done here.
  @classmethod
  def signature(cls, environment_spec: specs.EnvironmentSpec,
                extras_spec: types.NestedSpec = (),
                sequence_length: Optional[int] = None):
    """This is a helper method for generating signatures for Reverb tables.

    Signatures are useful for validating data types and shapes, see Reverb's
    documentation for details on how they are used.

    Args:
      environment_spec: A `specs.EnvironmentSpec` whose fields are nested
        structures with leaf nodes that have `.shape` and `.dtype` attributes.
        This should come from the environment that will be used to generate
        the data inserted into the Reverb table.
      extras_spec: A nested structure with leaf nodes that have `.shape` and
        `.dtype` attributes. The structure (and shapes/dtypes) of this must
        be the same as the `extras` passed into `ReverbAdder.add`.
      sequence_length: An optional integer representing the expected length of
        sequences that will be added to replay.

    Returns:
      A `Trajectory` whose leaf nodes are `tf.TensorSpec` objects.
    """

    def add_time_dim(paths: Iterable[str], spec: tf.TensorSpec):
      return tf.TensorSpec(shape=(sequence_length, *spec.shape),
                           dtype=spec.dtype,
                           name='/'.join(str(p) for p in paths))

    trajectory_env_spec, trajectory_extras_spec = tree.map_structure_with_path(
        add_time_dim, (environment_spec, extras_spec))

    spec_step = base.Trajectory(
        *trajectory_env_spec,
        start_of_episode=tf.TensorSpec(
            shape=(sequence_length,), dtype=tf.bool, name='start_of_episode'),
        extras=trajectory_extras_spec)

    return spec_step



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""R2D2 learner implementation."""

import functools
import time
from typing import Dict, Iterator, List, NamedTuple, Optional, Tuple

import acme
from acme.adders import reverb as adders
from acme.jax import networks as networks_lib
from acme.jax import utils
from acme.utils import async_utils
from acme.utils import counting
from acme.utils import loggers
import jax
import jax.numpy as jnp
import optax
import reverb
import rlax
import tree

_PMAP_AXIS_NAME = 'data'


class TrainingState(NamedTuple):
  """Holds the agent's training state."""
  params: networks_lib.Params
  target_params: networks_lib.Params
  opt_state: optax.OptState
  steps: int
  random_key: networks_lib.PRNGKey


class R2D2Learner(acme.Learner):
  """R2D2 learner."""

  def __init__(self,
               unroll: networks_lib.FeedForwardNetwork,
               initial_state: networks_lib.FeedForwardNetwork,
               batch_size: int,
               random_key: networks_lib.PRNGKey,
               burn_in_length: int,
               discount: float,
               importance_sampling_exponent: float,
               max_priority_weight: float,
               target_update_period: int,
               iterator: Iterator[reverb.ReplaySample],
               optimizer: optax.GradientTransformation,
               bootstrap_n: int = 5,
               tx_pair: rlax.TxPair = rlax.SIGNED_HYPERBOLIC_PAIR,
               clip_rewards: bool = False,
               max_abs_reward: float = 1.,
               use_core_state: bool = True,
               prefetch_size: int = 2,
               replay_client: Optional[reverb.Client] = None,
               counter: Optional[counting.Counter] = None,
               logger: Optional[loggers.Logger] = None):
    """Initializes the learner."""

    random_key, key_initial_1, key_initial_2 = jax.random.split(random_key, 3)
    initial_state_params = initial_state.init(key_initial_1, batch_size)
    initial_state = initial_state.apply(initial_state_params, key_initial_2,
                                        batch_size)

    def loss(
        params: networks_lib.Params,
        target_params: networks_lib.Params,
        key_grad: networks_lib.PRNGKey,
        sample: reverb.ReplaySample
    ) -> Tuple[jnp.ndarray, Tuple[jnp.ndarray, jnp.ndarray]]:
      """Computes mean transformed N-step loss for a batch of sequences."""

      # Convert sample data to sequence-major format [T, B, ...].
      data = utils.batch_to_sequence(sample.data)

      # Get core state & warm it up on observations for a burn-in period.
      if use_core_state:
        # Replay core state.
        online_state = jax.tree_map(lambda x: x[0], data.extras['core_state'])
      else:
        online_state = initial_state
      target_state = online_state

      # Maybe burn the core state in.
      if burn_in_length:
        burn_obs = jax.tree_map(lambda x: x[:burn_in_length], data.observation)
        key_grad, key1, key2 = jax.random.split(key_grad, 3)
        _, online_state = unroll.apply(params, key1, burn_obs, online_state)
        _, target_state = unroll.apply(target_params, key2, burn_obs,
                                       target_state)

      # Only get data to learn on from after the end of the burn in period.
      data = jax.tree_map(lambda seq: seq[burn_in_length:], data)

      # Unroll on sequences to get online and target Q-Values.
      key1, key2 = jax.random.split(key_grad)
      online_q, _ = unroll.apply(params, key1, data.observation, online_state)
      target_q, _ = unroll.apply(target_params, key2, data.observation,
                                 target_state)

      # Get value-selector actions from online Q-values for double Q-learning.
      selector_actions = jnp.argmax(online_q, axis=-1)
      # Preprocess discounts & rewards.
      discounts = (data.discount * discount).astype(online_q.dtype)
      rewards = data.reward
      if clip_rewards:
        rewards = jnp.clip(rewards, -max_abs_reward, max_abs_reward)
      rewards = rewards.astype(online_q.dtype)

      # Get N-step transformed TD error and loss.
      batch_td_error_fn = jax.vmap(
          functools.partial(
              rlax.transformed_n_step_q_learning,
              n=bootstrap_n,
              tx_pair=tx_pair),
          in_axes=1,
          out_axes=1)
      # TODO(b/183945808): when this bug is fixed, truncations of actions,
      # rewards, and discounts will no longer be necessary.
      batch_td_error = batch_td_error_fn(
          online_q[:-1],
          data.action[:-1],
          target_q[1:],
          selector_actions[1:],
          rewards[:-1],
          discounts[:-1])
      batch_loss = 0.5 * jnp.square(batch_td_error).sum(axis=0)

      # Importance weighting.
      probs = sample.info.probability
      importance_weights = (1. / (probs + 1e-6)).astype(online_q.dtype)
      importance_weights **= importance_sampling_exponent
      importance_weights /= jnp.max(importance_weights)
      mean_loss = jnp.mean(importance_weights * batch_loss)

      # Calculate priorities as a mixture of max and mean sequence errors.
      abs_td_error = jnp.abs(batch_td_error).astype(online_q.dtype)
      max_priority = max_priority_weight * jnp.max(abs_td_error, axis=0)
      mean_priority = (1 - max_priority_weight) * jnp.mean(abs_td_error, axis=0)
      priorities = (max_priority + mean_priority)

      return mean_loss, priorities

    def sgd_step(
        state: TrainingState,
        samples: reverb.ReplaySample
    ) -> Tuple[TrainingState, jnp.ndarray, Dict[str, jnp.ndarray]]:
      """Performs an update step, averaging over pmap replicas."""

      # Compute loss and gradients.
      grad_fn = jax.value_and_grad(loss, has_aux=True)
      key, key_grad = jax.random.split(state.random_key)
      (loss_value, priorities), gradients = grad_fn(state.params,
                                                    state.target_params,
                                                    key_grad,
                                                    samples)

      # Average gradients over pmap replicas before optimizer update.
      gradients = jax.lax.pmean(gradients, _PMAP_AXIS_NAME)

      # Apply optimizer updates.
      updates, new_opt_state = optimizer.update(gradients, state.opt_state)
      new_params = optax.apply_updates(state.params, updates)

      # Periodically update target networks.
      steps = state.steps + 1
      target_params = rlax.periodic_update(
          new_params, state.target_params, steps, self._target_update_period)

      new_state = TrainingState(
          params=new_params,
          target_params=target_params,
          opt_state=new_opt_state,
          steps=steps,
          random_key=key)
      return new_state, priorities, {'loss': loss_value}

    def update_priorities(
        keys_and_priorities: Tuple[jnp.ndarray, jnp.ndarray]):
      keys, priorities = keys_and_priorities
      keys, priorities = tree.map_structure(
          # Fetch array and combine device and batch dimensions.
          lambda x: utils.fetch_devicearray(x).reshape((-1,) + x.shape[2:]),
          (keys, priorities))
      replay_client.mutate_priorities(  # pytype: disable=attribute-error
          table=adders.DEFAULT_PRIORITY_TABLE,
          updates=dict(zip(keys, priorities)))

    # Internalise components, hyperparameters, logger, counter, and methods.
    self._replay_client = replay_client
    self._target_update_period = target_update_period
    self._counter = counter or counting.Counter()
    self._logger = logger or loggers.make_default_logger(
        'learner',
        asynchronous=True,
        serialize_fn=utils.fetch_devicearray,
        time_delta=1.)

    self._sgd_step = jax.pmap(sgd_step, axis_name=_PMAP_AXIS_NAME)
    self._async_priority_updater = async_utils.AsyncExecutor(update_priorities)

    # Initialise and internalise training state (parameters/optimiser state).
    random_key, key_init = jax.random.split(random_key)
    initial_params = unroll.init(key_init, initial_state)
    opt_state = optimizer.init(initial_params)

    state = TrainingState(
        params=initial_params,
        target_params=initial_params,
        opt_state=opt_state,
        steps=jnp.array(0),
        random_key=random_key)
    # Replicate parameters.
    self._state = utils.replicate_in_all_devices(state)

    # Shard multiple inputs with on-device prefetching.
    # We split samples in two outputs, the keys which need to be kept on-host
    # since int64 arrays are not supported in TPUs, and the entire sample
    # separately so it can be sent to the sgd_step method.
    def split_sample(sample: reverb.ReplaySample) -> utils.PrefetchingSplit:
      return utils.PrefetchingSplit(host=sample.info.key, device=sample)

    self._prefetched_iterator = utils.sharded_prefetch(
        iterator,
        buffer_size=prefetch_size,
        num_threads=jax.local_device_count(),
        split_fn=split_sample)

  def step(self):
    prefetching_split = next(self._prefetched_iterator)
    # The split_sample method passed to utils.sharded_prefetch specifies what
    # parts of the objects returned by the original iterator are kept in the
    # host and what parts are prefetched on-device.
    # In this case the host property of the prefetching split contains only the
    # replay keys and the device property is the prefetched full original
    # sample.
    keys, samples = prefetching_split.host, prefetching_split.device

    # Do a batch of SGD.
    start = time.time()
    self._state, priorities, metrics = self._sgd_step(self._state, samples)
    # Take metrics from first replica.
    metrics = utils.get_from_first_device(metrics)
    # Update our counts and record it.
    counts = self._counter.increment(steps=1, time_elapsed=time.time() - start)

    # Update priorities in replay.
    if self._replay_client:
      self._async_priority_updater.put((keys, priorities))

    # Attempt to write logs.
    self._logger.write({**metrics, **counts})

  def get_variables(self, names: List[str]) -> List[networks_lib.Params]:
    # Return first replica of parameters.
    return [utils.get_from_first_device(self._state.params)]

  def save(self) -> TrainingState:
    # Serialize only the first replica of parameters and optimizer state.
    return jax.tree_map(utils.get_from_first_device, self._state)

  def restore(self, state: TrainingState):
    self._state = utils.replicate_in_all_devices(state)



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Networks used for handling illegal actions."""

from typing import Any, Callable, Iterable, Optional, Union

# pytype: disable=import-error
from acme.wrappers import open_spiel_wrapper
# pytype: enable=import-error

import numpy as np
import sonnet as snt
import tensorflow as tf
import tensorflow_probability as tfp

tfd = tfp.distributions


class MaskedSequential(snt.Module):
  """Applies a legal actions mask to a linear chain of modules / callables.

  It is assumed the trailing dimension of the final layer (representing
  action values) is the same as the trailing dimension of legal_actions.
  """

  def __init__(self,
               layers: Optional[Iterable[Callable[..., Any]]] = None,
               name: str = 'MaskedSequential'):
    super().__init__(name=name)
    self._layers = list(layers) if layers is not None else []
    self._illegal_action_penalty = -1e9
    # Note: illegal_action_penalty cannot be -np.inf because trfl's qlearning
    # ops utilize a batched_index function that returns NaN whenever -np.inf
    # is present among action values.

  def __call__(self, inputs: open_spiel_wrapper.OLT) -> tf.Tensor:
    # Extract observation, legal actions, and terminal
    outputs = inputs.observation
    legal_actions = inputs.legal_actions
    terminal = inputs.terminal

    for mod in self._layers:
      outputs = mod(outputs)

    # Apply legal actions mask
    outputs = tf.where(tf.equal(legal_actions, 1), outputs,
                       tf.fill(tf.shape(outputs), self._illegal_action_penalty))

    # When computing the Q-learning target (r_t + d_t * max q_t) we need to
    # ensure max q_t = 0 in terminal states.
    outputs = tf.where(tf.equal(terminal, 1), tf.zeros_like(outputs), outputs)

    return outputs


# FIXME: Add functionality to support decaying epsilon parameter.
# FIXME: This is a modified version of trfl's epsilon_greedy() which
# incorporates code from the bug fix described here
# https://github.com/deepmind/trfl/pull/28
class EpsilonGreedy(snt.Module):
  """Computes an epsilon-greedy distribution over actions.

  This policy does the following:
  - With probability 1 - epsilon, take the action corresponding to the highest
  action value, breaking ties uniformly at random.
  - With probability epsilon, take an action uniformly at random.
  """

  def __init__(self,
               epsilon: Union[tf.Tensor, float],
               threshold: float,
               name: str = 'EpsilonGreedy'):
    """Initialize the policy.

    Args:
      epsilon: Exploratory param with value between 0 and 1.
      threshold: Action values must exceed this value to qualify as a legal
        action and possibly be selected by the policy.
      name: Name of the network.

    Returns:
      policy: tfp.distributions.Categorical distribution representing the
        policy.
    """
    super().__init__(name=name)
    self._epsilon = tf.Variable(epsilon, trainable=False)
    self._threshold = threshold

  def __call__(self, action_values: tf.Tensor) -> tfd.Categorical:
    legal_actions_mask = tf.where(
        tf.math.less_equal(action_values, self._threshold),
        tf.fill(tf.shape(action_values), 0.),
        tf.fill(tf.shape(action_values), 1.))

    # Dithering action distribution.
    dither_probs = 1 / tf.reduce_sum(legal_actions_mask, axis=-1,
                                     keepdims=True) * legal_actions_mask
    masked_action_values = tf.where(tf.equal(legal_actions_mask, 1),
                                    action_values,
                                    tf.fill(tf.shape(action_values), -np.inf))
    # Greedy action distribution, breaking ties uniformly at random.
    max_value = tf.reduce_max(masked_action_values, axis=-1, keepdims=True)
    greedy_probs = tf.cast(
        tf.equal(action_values * legal_actions_mask, max_value),
        action_values.dtype)

    greedy_probs /= tf.reduce_sum(greedy_probs, axis=-1, keepdims=True)

    # Epsilon-greedy action distribution.
    probs = self._epsilon * dither_probs + (1 - self._epsilon) * greedy_probs

    # Make the policy object.
    policy = tfd.Categorical(probs=probs)

    return policy



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for IMPALA agent."""

from absl.testing import absltest
import acme
from acme import specs
from acme.agents.tf import impala
from acme.testing import fakes
from acme.tf import networks
import numpy as np
import sonnet as snt


def _make_network(action_spec: specs.DiscreteArray) -> snt.RNNCore:
  return snt.DeepRNN([
      snt.Flatten(),
      snt.LSTM(20),
      snt.nets.MLP([50, 50]),
      networks.PolicyValueHead(action_spec.num_values),
  ])


class IMPALATest(absltest.TestCase):

  # TODO(b/200509080): This test case is timing out.
  @absltest.SkipTest
  def test_impala(self):
    # Create a fake environment to test with.
    environment = fakes.DiscreteEnvironment(
        num_actions=5,
        num_observations=10,
        obs_dtype=np.float32,
        episode_length=10)
    spec = specs.make_environment_spec(environment)

    # Construct the agent.
    agent = impala.IMPALA(
        environment_spec=spec,
        network=_make_network(spec.actions),
        sequence_length=3,
        sequence_period=3,
        batch_size=6,
    )

    # Try running the environment loop. We have no assertions here because all
    # we care about is that the agent runs without raising any errors.
    loop = acme.EnvironmentLoop(environment, agent)
    loop.run(num_episodes=20)


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wrapper that implements environment step limit."""

from typing import Optional
from acme import types
from acme.wrappers import base
import dm_env


class StepLimitWrapper(base.EnvironmentWrapper):
  """A wrapper which truncates episodes at the specified step limit."""

  def __init__(self, environment: dm_env.Environment,
               step_limit: Optional[int] = None):
    super().__init__(environment)
    self._step_limit = step_limit
    self._elapsed_steps = 0

  def reset(self) -> dm_env.TimeStep:
    self._elapsed_steps = 0
    return self._environment.reset()

  def step(self, action: types.NestedArray) -> dm_env.TimeStep:
    timestep = self._environment.step(action)
    self._elapsed_steps += 1
    if self._step_limit is not None and self._elapsed_steps >= self._step_limit:
      return dm_env.truncation(
          timestep.reward, timestep.observation, timestep.discount)
    return timestep



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SVG0 actor implementation."""

from typing import Optional

from acme import adders
from acme import types

from acme.agents.tf import actors
from acme.tf import utils as tf2_utils
from acme.tf import variable_utils as tf2_variable_utils

import dm_env
import sonnet as snt


class SVG0Actor(actors.FeedForwardActor):
  """An actor that also returns `log_prob`."""

  def __init__(
      self,
      policy_network: snt.Module,
      adder: Optional[adders.Adder] = None,
      variable_client: Optional[tf2_variable_utils.VariableClient] = None,
      deterministic_policy: Optional[bool] = False,
  ):
    super().__init__(policy_network, adder, variable_client)
    self._log_prob = None
    self._deterministic_policy = deterministic_policy

  def select_action(self, observation: types.NestedArray) -> types.NestedArray:
    # Add a dummy batch dimension and as a side effect convert numpy to TF.
    batched_observation = tf2_utils.add_batch_dim(observation)

    # Compute the policy, conditioned on the observation.
    policy = self._policy_network(batched_observation)
    if self._deterministic_policy:
      action = policy.mean()
    else:
      action = policy.sample()
    self._log_prob = policy.log_prob(action)
    return tf2_utils.to_numpy_squeeze(action)

  def observe(
      self,
      action: types.NestedArray,
      next_timestep: dm_env.TimeStep,
  ):
    if not self._adder:
      return

    extras = {'log_prob': self._log_prob}
    extras = tf2_utils.to_numpy_squeeze(extras)
    self._adder.add(action, next_timestep, extras)



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Episode adders.

This implements full episode adders, potentially with padding.
"""

from typing import Callable, Optional, Iterable, Tuple

from acme import specs
from acme import types
from acme.adders.reverb import base
from acme.adders.reverb import utils

import dm_env
import numpy as np
import reverb
import tensorflow as tf
import tree

_PaddingFn = Callable[[Tuple[int, ...], np.dtype], np.ndarray]


class EpisodeAdder(base.ReverbAdder):
  """Adder which adds entire episodes as trajectories."""

  def __init__(
      self,
      client: reverb.Client,
      max_sequence_length: int,
      delta_encoded: bool = False,
      priority_fns: Optional[base.PriorityFnMapping] = None,
      max_in_flight_items: int = 1,
      padding_fn: Optional[_PaddingFn] = None,
      # Deprecated kwargs.
      chunk_length: Optional[int] = None,
  ):
    del chunk_length

    super().__init__(
        client=client,
        max_sequence_length=max_sequence_length,
        delta_encoded=delta_encoded,
        priority_fns=priority_fns,
        max_in_flight_items=max_in_flight_items,
    )
    self._padding_fn = padding_fn

  def add(
      self,
      action: types.NestedArray,
      next_timestep: dm_env.TimeStep,
      extras: types.NestedArray = (),
  ):
    if self._writer.episode_steps >= self._max_sequence_length - 1:
      raise ValueError(
          'The number of observations within the same episode will exceed '
          'max_sequence_length with the addition of this transition.')

    super().add(action, next_timestep, extras)

  def _write(self):
    # This adder only writes at the end of the episode, see _write_last()
    pass

  def _write_last(self):
    if self._padding_fn is not None and self._writer.episode_steps < self._max_sequence_length:
      history = self._writer.history
      padding_step = dict(
          observation=history['observation'],
          action=history['action'],
          reward=history['reward'],
          discount=history['discount'],
          extras=history.get('extras', ()))
      # Get shapes and dtypes from the last element.
      padding_step = tree.map_structure(
          lambda col: self._padding_fn(col[-1].shape, col[-1].dtype),
          padding_step)
      padding_step['start_of_episode'] = False
      while self._writer.episode_steps < self._max_sequence_length:
        self._writer.append(padding_step)

    trajectory = tree.map_structure(lambda x: x[:], self._writer.history)

    # Pack the history into a base.Step structure and get numpy converted
    # variant for priotiy computation.
    trajectory = base.Trajectory(**trajectory)

    # Calculate the priority for this episode.
    table_priorities = utils.calculate_priorities(self._priority_fns,
                                                  trajectory)

    # Create a prioritized item for each table.
    for table_name, priority in table_priorities.items():
      self._writer.create_item(table_name, priority, trajectory)
      self._writer.flush(self._max_in_flight_items)

  # TODO(b/185309817): make this into a standalone method.
  @classmethod
  def signature(cls,
                environment_spec: specs.EnvironmentSpec,
                extras_spec: types.NestedSpec = (),
                sequence_length: Optional[int] = None):
    """This is a helper method for generating signatures for Reverb tables.

    Signatures are useful for validating data types and shapes, see Reverb's
    documentation for details on how they are used.

    Args:
      environment_spec: A `specs.EnvironmentSpec` whose fields are nested
        structures with leaf nodes that have `.shape` and `.dtype` attributes.
        This should come from the environment that will be used to generate the
        data inserted into the Reverb table.
      extras_spec: A nested structure with leaf nodes that have `.shape` and
        `.dtype` attributes. The structure (and shapes/dtypes) of this must be
        the same as the `extras` passed into `ReverbAdder.add`.
      sequence_length: An optional integer representing the expected length of
        sequences that will be added to replay.

    Returns:
      A `Step` whose leaf nodes are `tf.TensorSpec` objects.
    """

    def add_time_dim(paths: Iterable[str], spec: tf.TensorSpec):
      return tf.TensorSpec(
          shape=(sequence_length, *spec.shape),
          dtype=spec.dtype,
          name='/'.join(str(p) for p in paths))

    trajectory_env_spec, trajectory_extras_spec = tree.map_structure_with_path(
        add_time_dim, (environment_spec, extras_spec))

    trajectory_spec = base.Trajectory(
        *trajectory_env_spec,
        start_of_episode=tf.TensorSpec(
            shape=(sequence_length,), dtype=tf.bool, name='start_of_episode'),
        extras=trajectory_extras_spec)

    return trajectory_spec



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TD3 networks definition."""
import dataclasses
from typing import Callable, Sequence

from acme import specs
from acme import types
from acme.jax import networks as networks_lib
from acme.jax import utils
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np


@dataclasses.dataclass
class TD3Networks:
  """Network and pure functions for the TD3 agent."""
  policy_network: networks_lib.FeedForwardNetwork
  critic_network: networks_lib.FeedForwardNetwork
  twin_critic_network: networks_lib.FeedForwardNetwork
  add_policy_noise: Callable[[types.NestedArray, networks_lib.PRNGKey,
                              float, float], types.NestedArray]


def get_default_behavior_policy(networks: TD3Networks,
                                action_specs: specs.BoundedArray,
                                sigma: float):
  """Selects action according to the policy."""
  def behavior_policy(params: networks_lib.Params, key: networks_lib.PRNGKey,
                      observation: types.NestedArray):
    action = networks.policy_network.apply(params, observation)
    noise = jax.random.normal(key, shape=action.shape) * sigma
    noisy_action = jnp.clip(action + noise,
                            action_specs.minimum, action_specs.maximum)
    return noisy_action
  return behavior_policy


def make_networks(
    spec: specs.EnvironmentSpec,
    hidden_layer_sizes: Sequence[int] = (256, 256)) -> TD3Networks:
  """Creates networks used by the agent.

  The networks used are based on LayerNormMLP, which is different than the
  MLP with relu activation described in TD3 (which empirically performs worse).

  Args:
    spec: Environment specs
    hidden_layer_sizes: list of sizes of hidden layers in actor/critic networks

  Returns:
    network: TD3Networks
  """

  action_specs = spec.actions
  num_dimensions = np.prod(action_specs.shape, dtype=int)

  def add_policy_noise(action: types.NestedArray,
                       key: networks_lib.PRNGKey,
                       target_sigma: float,
                       noise_clip: float) -> types.NestedArray:
    """Adds action noise to bootstrapped Q-value estimate in critic loss."""
    noise = jax.random.normal(key=key, shape=action_specs.shape) * target_sigma
    noise = jnp.clip(noise, -noise_clip, noise_clip)
    return jnp.clip(action + noise, action_specs.minimum, action_specs.maximum)

  def _actor_fn(obs: types.NestedArray) -> types.NestedArray:
    network = hk.Sequential([
        networks_lib.LayerNormMLP(hidden_layer_sizes,
                                  activate_final=True),
        networks_lib.NearZeroInitializedLinear(num_dimensions),
        networks_lib.TanhToSpec(spec.actions),
    ])
    return network(obs)

  def _critic_fn(obs: types.NestedArray,
                 action: types.NestedArray) -> types.NestedArray:
    network1 = hk.Sequential([
        networks_lib.LayerNormMLP(list(hidden_layer_sizes) + [1]),
    ])
    input_ = jnp.concatenate([obs, action], axis=-1)
    value = network1(input_)
    return jnp.squeeze(value)

  policy = hk.without_apply_rng(hk.transform(_actor_fn))
  critic = hk.without_apply_rng(hk.transform(_critic_fn))

  # Create dummy observations and actions to create network parameters.
  dummy_action = utils.zeros_like(spec.actions)
  dummy_obs = utils.zeros_like(spec.observations)
  dummy_action = utils.add_batch_dim(dummy_action)
  dummy_obs = utils.add_batch_dim(dummy_obs)

  network = TD3Networks(
      policy_network=networks_lib.FeedForwardNetwork(
          lambda key: policy.init(key, dummy_obs), policy.apply),
      critic_network=networks_lib.FeedForwardNetwork(
          lambda key: critic.init(key, dummy_obs, dummy_action), critic.apply),
      twin_critic_network=networks_lib.FeedForwardNetwork(
          lambda key: critic.init(key, dummy_obs, dummy_action), critic.apply),
      add_policy_noise=add_policy_noise)

  return network



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SgdLearner takes steps of SGD on a LossFn."""

import functools
import time
from typing import Dict, Iterator, List, NamedTuple, Optional, Tuple

import acme
from acme.adders import reverb as adders
from acme.jax import networks as networks_lib
from acme.jax import utils
from acme.utils import async_utils
from acme.utils import counting
from acme.utils import loggers
import jax
import jax.numpy as jnp
import optax
import reverb
import rlax
import tree
import typing_extensions


class ReverbUpdate(NamedTuple):
  """Tuple for updating reverb priority information."""
  keys: jnp.ndarray
  priorities: jnp.ndarray


class LossExtra(NamedTuple):
  """Extra information that is returned along with loss value."""
  metrics: Dict[str, jnp.DeviceArray]
  reverb_update: Optional[ReverbUpdate] = None


class LossFn(typing_extensions.Protocol):
  """A LossFn calculates a loss on a single batch of data."""

  def __call__(self,
               network: networks_lib.FeedForwardNetwork,
               params: networks_lib.Params,
               target_params: networks_lib.Params,
               batch: reverb.ReplaySample,
               key: networks_lib.PRNGKey) -> Tuple[jnp.DeviceArray, LossExtra]:
    """Calculates a loss on a single batch of data."""


class TrainingState(NamedTuple):
  """Holds the agent's training state."""
  params: networks_lib.Params
  target_params: networks_lib.Params
  opt_state: optax.OptState
  steps: int
  rng_key: networks_lib.PRNGKey


class SGDLearner(acme.Learner):
  """An Acme learner based around SGD on batches.

  This learner currently supports optional prioritized replay and assumes a
  TrainingState as described above.
  """

  def __init__(self,
               network: networks_lib.FeedForwardNetwork,
               loss_fn: LossFn,
               optimizer: optax.GradientTransformation,
               data_iterator: Iterator[reverb.ReplaySample],
               target_update_period: int,
               random_key: networks_lib.PRNGKey,
               replay_client: Optional[reverb.Client] = None,
               replay_table_name: str = adders.DEFAULT_PRIORITY_TABLE,
               counter: Optional[counting.Counter] = None,
               logger: Optional[loggers.Logger] = None,
               num_sgd_steps_per_step: int = 1):
    """Initialize the SGD learner."""
    self.network = network

    # Internalize the loss_fn with network.
    self._loss = jax.jit(functools.partial(loss_fn, self.network))

    # SGD performs the loss, optimizer update and periodic target net update.
    def sgd_step(state: TrainingState,
                 batch: reverb.ReplaySample) -> Tuple[TrainingState, LossExtra]:
      next_rng_key, rng_key = jax.random.split(state.rng_key)
      # Implements one SGD step of the loss and updates training state
      (loss, extra), grads = jax.value_and_grad(self._loss, has_aux=True)(
          state.params, state.target_params, batch, rng_key)
      extra.metrics.update({'total_loss': loss})

      # Apply the optimizer updates
      updates, new_opt_state = optimizer.update(grads, state.opt_state)
      new_params = optax.apply_updates(state.params, updates)

      # Periodically update target networks.
      steps = state.steps + 1
      target_params = rlax.periodic_update(
          new_params, state.target_params, steps, target_update_period)
      new_training_state = TrainingState(
          new_params, target_params, new_opt_state, steps, next_rng_key)
      return new_training_state, extra

    def postprocess_aux(extra: LossExtra) -> LossExtra:
      reverb_update = jax.tree_map(lambda a: jnp.reshape(a, (-1, *a.shape[2:])),
                                   extra.reverb_update)
      return extra._replace(
          metrics=jax.tree_map(jnp.mean, extra.metrics),
          reverb_update=reverb_update)

    self._num_sgd_steps_per_step = num_sgd_steps_per_step
    sgd_step = utils.process_multiple_batches(sgd_step, num_sgd_steps_per_step,
                                              postprocess_aux)
    self._sgd_step = jax.jit(sgd_step)

    # Internalise agent components
    self._data_iterator = utils.prefetch(data_iterator)
    self._target_update_period = target_update_period
    self._counter = counter or counting.Counter()
    self._logger = logger or loggers.TerminalLogger('learner', time_delta=1.)

    # Do not record timestamps until after the first learning step is done.
    # This is to avoid including the time it takes for actors to come online and
    # fill the replay buffer.
    self._timestamp = None

    # Initialize the network parameters
    key_params, key_target, key_state = jax.random.split(random_key, 3)
    initial_params = self.network.init(key_params)
    initial_target_params = self.network.init(key_target)
    self._state = TrainingState(
        params=initial_params,
        target_params=initial_target_params,
        opt_state=optimizer.init(initial_params),
        steps=0,
        rng_key=key_state,
    )

    # Update replay priorities
    def update_priorities(reverb_update: ReverbUpdate) -> None:
      if replay_client is None:
        return
      keys, priorities = tree.map_structure(
          utils.fetch_devicearray,
          (reverb_update.keys, reverb_update.priorities))
      replay_client.mutate_priorities(
          table=replay_table_name,
          updates=dict(zip(keys, priorities)))
    self._replay_client = replay_client
    self._async_priority_updater = async_utils.AsyncExecutor(update_priorities)

  def step(self):
    """Takes one SGD step on the learner."""
    with jax.profiler.StepTraceAnnotation('step',
                                          step_num=self._state.steps):
      batch = next(self._data_iterator)
      self._state, extra = self._sgd_step(self._state, batch)

      # Compute elapsed time.
      timestamp = time.time()
      elapsed = timestamp - self._timestamp if self._timestamp else 0
      self._timestamp = timestamp

      if self._replay_client and extra.reverb_update:
        reverb_update = extra.reverb_update._replace(keys=batch.info.key)
        self._async_priority_updater.put(reverb_update)

      steps_per_sec = (self._num_sgd_steps_per_step / elapsed) if elapsed else 0
      extra.metrics['steps_per_second'] = steps_per_sec

      # Update our counts and record it.
      result = self._counter.increment(
          steps=self._num_sgd_steps_per_step, walltime=elapsed)
      result.update(extra.metrics)
      self._logger.write(result)

  def get_variables(self, names: List[str]) -> List[networks_lib.Params]:
    return [self._state.params]

  def save(self) -> TrainingState:
    return self._state

  def restore(self, state: TrainingState):
    self._state = state



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DQN learner implementation."""

from typing import Iterator, Optional

from acme.adders import reverb as adders
from acme.agents.jax.dqn import learning_lib
from acme.agents.jax.dqn import losses
from acme.jax import networks as networks_lib
from acme.utils import counting
from acme.utils import loggers
import optax
import reverb


class DQNLearner(learning_lib.SGDLearner):
  """DQN learner.

  We are in the process of migrating towards a more general SGDLearner to allow
  for easy configuration of the loss. This is maintained now for compatibility.
  """

  def __init__(self,
               network: networks_lib.FeedForwardNetwork,
               discount: float,
               importance_sampling_exponent: float,
               target_update_period: int,
               iterator: Iterator[reverb.ReplaySample],
               optimizer: optax.GradientTransformation,
               random_key: networks_lib.PRNGKey,
               max_abs_reward: float = 1.,
               huber_loss_parameter: float = 1.,
               replay_client: Optional[reverb.Client] = None,
               replay_table_name: str = adders.DEFAULT_PRIORITY_TABLE,
               counter: Optional[counting.Counter] = None,
               logger: Optional[loggers.Logger] = None):
    """Initializes the learner."""
    loss_fn = losses.PrioritizedDoubleQLearning(
        discount=discount,
        importance_sampling_exponent=importance_sampling_exponent,
        max_abs_reward=max_abs_reward,
        huber_loss_parameter=huber_loss_parameter,
    )
    super().__init__(
        network=network,
        loss_fn=loss_fn,
        optimizer=optimizer,
        data_iterator=iterator,
        target_update_period=target_update_period,
        random_key=random_key,
        replay_client=replay_client,
        replay_table_name=replay_table_name,
        counter=counter,
        logger=logger,
    )



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Local agent based on builders."""

from typing import Any, Optional

from acme import specs
from acme.agents import agent
from acme.agents.jax import builders
from acme.jax import utils
from acme.tf import savers
from acme.utils import counting
import jax
import reverb


class LocalLayout(agent.Agent):
  """An Agent that runs an algorithm defined by 'builder' on a single machine.
  """

  def __init__(
      self,
      seed: int,
      environment_spec: specs.EnvironmentSpec,
      builder: builders.GenericActorLearnerBuilder,
      networks: Any,
      policy_network: Any,
      workdir: Optional[str] = '~/acme',
      min_replay_size: int = 1000,
      samples_per_insert: float = 256.0,
      batch_size: int = 256,
      num_sgd_steps_per_step: int = 1,
      prefetch_size: int = 1,
      device_prefetch: bool = True,
      counter: Optional[counting.Counter] = None,
      checkpoint: bool = True,
  ):
    """Initialize the agent.

    Args:
      seed: A random seed to use for this layout instance.
      environment_spec: description of the actions, observations, etc.
      builder: builder defining an RL algorithm to train.
      networks: network objects to be passed to the learner.
      policy_network: function that given an observation returns actions.
      workdir: if provided saves the state of the learner and the counter
        (if the counter is not None) into workdir.
      min_replay_size: minimum replay size before updating.
      samples_per_insert: number of samples to take from replay for every insert
        that is made.
      batch_size: batch size for updates.
      num_sgd_steps_per_step: how many sgd steps a learner does per 'step' call.
        For performance reasons (especially to reduce TPU host-device transfer
        times) it is performance-beneficial to do multiple sgd updates at once,
        provided that it does not hurt the training, which needs to be verified
        empirically for each environment.
      prefetch_size: whether to prefetch iterator.
      device_prefetch: whether prefetching should happen to a device.
      counter: counter object used to keep track of steps.
      checkpoint: boolean indicating whether to checkpoint the learner
        and the counter (if the counter is not None).
    """
    if prefetch_size < 0:
      raise ValueError(f'Prefetch size={prefetch_size} should be non negative')

    key = jax.random.PRNGKey(seed)

    # Create the replay server and grab its address.
    replay_tables = builder.make_replay_tables(environment_spec)
    replay_server = reverb.Server(replay_tables, port=None)
    replay_client = reverb.Client(f'localhost:{replay_server.port}')

    # Create actor, dataset, and learner for generating, storing, and consuming
    # data respectively.
    adder = builder.make_adder(replay_client)

    def _is_reverb_queue(reverb_table: reverb.Table,
                         reverb_client: reverb.Client) -> bool:
      """Returns True iff the Reverb Table is actually a queue."""
      # TODO(sinopalnikov): make it more generic and check for a table that
      # needs special handling on update.
      info = reverb_client.server_info()
      table_info = info[reverb_table.name]
      is_queue = (
          table_info.max_times_sampled == 1 and
          table_info.sampler_options.fifo and
          table_info.remover_options.fifo)
      return is_queue

    is_reverb_queue = any(_is_reverb_queue(table, replay_client)
                          for table in replay_tables)

    dataset = builder.make_dataset_iterator(replay_client)
    if prefetch_size > 1:
      device = jax.devices()[0] if device_prefetch else None
      dataset = utils.prefetch(dataset, buffer_size=prefetch_size,
                               device=device)
    learner_key, key = jax.random.split(key)
    learner = builder.make_learner(
        random_key=learner_key,
        networks=networks,
        dataset=dataset,
        replay_client=replay_client,
        counter=counter)
    if not checkpoint or workdir is None:
      self._checkpointer = None
    else:
      objects_to_save = {'learner': learner}
      if counter is not None:
        objects_to_save.update({'counter': counter})
      self._checkpointer = savers.Checkpointer(
          objects_to_save,
          time_delta_minutes=30,
          subdirectory='learner',
          directory=workdir,
          add_uid=(workdir == '~/acme'))

    actor_key, key = jax.random.split(key)
    actor = builder.make_actor(
        actor_key, policy_network, adder, variable_source=learner)
    self._custom_update_fn = None
    if is_reverb_queue:
      # Reverb queue requires special handling on update: custom logic to
      # decide when it is safe to make a learner step. This is only needed for
      # the local agent, where the actor and the learner are running
      # synchronously and the learner will deadlock if it makes a step with
      # no data available.
      def custom_update():
        should_update_actor = False
        # Run a number of learner steps (usually gradient steps).
        # TODO(raveman): This is wrong. When running multi-level learners,
        # different levels might have different batch sizes. Find a solution.
        while all(table.can_sample(batch_size) for table in replay_tables):
          learner.step()
          should_update_actor = True

        if should_update_actor:
          # "wait=True" to make it more onpolicy
          actor.update(wait=True)

      self._custom_update_fn = custom_update

    effective_batch_size = batch_size * num_sgd_steps_per_step
    super().__init__(
        actor=actor,
        learner=learner,
        min_observations=max(effective_batch_size, min_replay_size),
        observations_per_step=float(effective_batch_size) / samples_per_insert)

    # Save the replay so we don't garbage collect it.
    self._replay_server = replay_server

  def update(self):
    if self._custom_update_fn:
      self._custom_update_fn()
    else:
      super().update()
    if self._checkpointer:
      self._checkpointer.save()

  def save(self):
    """Checkpoint the state of the agent."""
    if self._checkpointer:
      self._checkpointer.save(force=True)



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TD3 Builder."""
from typing import Callable, Iterator, List, Optional

from acme import adders
from acme import core
from acme import specs
from acme.adders import reverb as adders_reverb
from acme.agents.jax import actor_core as actor_core_lib
from acme.agents.jax import actors
from acme.agents.jax import builders
from acme.agents.jax.td3 import config as td3_config
from acme.agents.jax.td3 import learning
from acme.agents.jax.td3 import networks as td3_networks
from acme.datasets import reverb as datasets
from acme.jax import networks as networks_lib
from acme.jax import variable_utils
from acme.utils import counting
from acme.utils import loggers
import optax
import reverb
from reverb import rate_limiters


class TD3Builder(builders.ActorLearnerBuilder):
  """TD3 Builder."""

  def __init__(
      self,
      config: td3_config.TD3Config,
      logger_fn: Callable[[], loggers.Logger] = lambda: None,
  ):
    """Creates a TD3 learner, a behavior policy and an eval actor.

    Args:
      config: a config with TD3 hps
      logger_fn: a logger factory for the learner
    """
    self._config = config
    self._logger_fn = logger_fn

  def make_learner(
      self,
      random_key: networks_lib.PRNGKey,
      networks: td3_networks.TD3Networks,
      dataset: Iterator[reverb.ReplaySample],
      replay_client: Optional[reverb.Client] = None,
      counter: Optional[counting.Counter] = None,
  ) -> core.Learner:

    critic_optimizer = optax.adam(self._config.critic_learning_rate)
    twin_critic_optimizer = optax.adam(self._config.critic_learning_rate)
    policy_optimizer = optax.adam(self._config.policy_learning_rate)

    if self._config.policy_gradient_clipping is not None:
      policy_optimizer = optax.chain(
          optax.clip_by_global_norm(self._config.policy_gradient_clipping),
          policy_optimizer)

    return learning.TD3Learner(
        networks=networks,
        random_key=random_key,
        discount=self._config.discount,
        target_sigma=self._config.target_sigma,
        noise_clip=self._config.noise_clip,
        policy_optimizer=policy_optimizer,
        critic_optimizer=critic_optimizer,
        twin_critic_optimizer=twin_critic_optimizer,
        num_sgd_steps_per_step=self._config.num_sgd_steps_per_step,
        iterator=dataset,
        logger=self._logger_fn(),
        counter=counter)

  def make_actor(
      self,
      random_key: networks_lib.PRNGKey,
      policy_network,
      adder: Optional[adders.Adder] = None,
      variable_source: Optional[core.VariableSource] = None,
  ) -> core.Actor:
    assert variable_source is not None
    actor_core = actor_core_lib.batched_feed_forward_to_actor_core(
        policy_network)
    # Inference happens on CPU, so it's better to move variables there too.
    variable_client = variable_utils.VariableClient(variable_source, 'policy',
                                                    device='cpu')
    return actors.GenericActor(
        actor_core, random_key, variable_client, adder, backend='cpu')

  def make_replay_tables(
      self, environment_spec: specs.EnvironmentSpec) -> List[reverb.Table]:
    """Creates reverb tables for the algorithm."""
    samples_per_insert_tolerance = (
        self._config.samples_per_insert_tolerance_rate *
        self._config.samples_per_insert)
    error_buffer = self._config.min_replay_size * samples_per_insert_tolerance
    limiter = rate_limiters.SampleToInsertRatio(
        min_size_to_sample=self._config.min_replay_size,
        samples_per_insert=self._config.samples_per_insert,
        error_buffer=error_buffer)
    return [reverb.Table(
        name=self._config.replay_table_name,
        sampler=reverb.selectors.Uniform(),
        remover=reverb.selectors.Fifo(),
        max_size=self._config.max_replay_size,
        rate_limiter=limiter,
        signature=adders_reverb.NStepTransitionAdder.signature(
            environment_spec))]

  def make_dataset_iterator(
      self, replay_client: reverb.Client) -> Iterator[reverb.ReplaySample]:
    """Creates a dataset iterator to use for learning."""
    dataset = datasets.make_reverb_dataset(
        table=self._config.replay_table_name,
        server_address=replay_client.server_address,
        batch_size=(
            self._config.batch_size * self._config.num_sgd_steps_per_step),
        prefetch_size=self._config.prefetch_size,
        transition_adder=True)
    return dataset.as_numpy_iterator()

  def make_adder(self, replay_client: reverb.Client) -> adders.Adder:
    """Creates an adder which handles observations."""
    return adders_reverb.NStepTransitionAdder(
        priority_fns={self._config.replay_table_name: None},
        client=replay_client,
        n_step=self._config.n_step,
        discount=self._config.discount)
